import os.path

import pytest
import glob

from refnx.dataset import ReflectDataset, Data1D, load_data, OrsoDataset
from refnx.dataset.reflectdataset import load_orso
import numpy as np
from numpy.testing import assert_equal


class TestReflectDataset:
    @pytest.fixture(autouse=True)
    def setup_method(self, tmpdir):
        self.pth = os.path.dirname(os.path.abspath(__file__))
        self.cwd = os.getcwd()
        self.tmpdir = tmpdir.strpath
        os.chdir(self.tmpdir)

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_ort_load(self):
        f = os.path.join(self.pth, "ORSO_data.ort")
        try:
            load_orso(f)
        except ImportError:
            # load_orso had problems on Python 3.10, so bypass the test
            return

        d = load_data(f)
        assert len(d) == 2
        assert isinstance(d, OrsoDataset)
        d.refresh()

    def test_load_data(self):
        # test the load_data function by trying to load all the files in the
        # test directory
        fs = glob.glob("*.*")
        fs = [f for f in fs if not f.endswith(".py")]
        fs = [f for f in fs if not f.startswith("coef_")]

        for f in fs:
            load_data(f)



import os
from os.path import join as pjoin
import numbers
import warnings
import pickle

import pytest
import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_,
    assert_equal,
    assert_array_less,
    assert_allclose,
)
import h5py

import refnx.reduce.platypusnexus as plp
from refnx.reduce import (
    PlatypusReduce,
    PlatypusNexus,
    SpatzNexus,
    basename_datafile,
    catalogue,
    SpinChannel,
    SpinSet,
)
from refnx.reduce.peak_utils import gauss
from refnx.reduce.platypusnexus import (
    fore_back_region,
    EXTENT_MULT,
    PIXEL_OFFSET,
    create_detector_norm,
    Catalogue,
    PlatypusCatalogue,
)


class TestSpinSet(object):
    @pytest.mark.usefixtures("no_data_directory")
    @pytest.fixture(autouse=True)
    def setup_method(self, tmpdir, data_directory):
        self.pth = pjoin(data_directory, "reduce", "PNR_files")

        def fpath(f):
            return pjoin(self.pth, f)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)

            self.spinset = SpinSet(
                down_down=fpath("PLP0008864.nx.hdf"),
                up_up=fpath("PLP0008861.nx.hdf"),
                down_up=fpath("PLP0008863.nx.hdf"),
                up_down=fpath("PLP0008862.nx.hdf"),
            )

            self.spinset_3 = SpinSet(
                down_down=fpath("PLP0008864.nx.hdf"),
                up_up=fpath("PLP0008861.nx.hdf"),
                down_up=fpath("PLP0008863.nx.hdf"),
            )
            self.spinset_2 = SpinSet(
                down_down=fpath("PLP0008864.nx.hdf"),
                up_up=fpath("PLP0008861.nx.hdf"),
            )

        self.cwd = os.getcwd()

        self.tmpdir = tmpdir.strpath
        os.chdir(self.tmpdir)
        return 0

    def test_spin_channels(self):
        assert self.spinset.spin_channels == [(0, 0), (0, 1), (1, 0), (1, 1)]
        assert self.spinset_3.spin_channels == [(0, 0), (0, 1), None, (1, 1)]
        assert self.spinset_2.spin_channels == [(0, 0), None, None, (1, 1)]

    def test_process(self):
        self.spinset.process()
        assert self.spinset.dd.processed_spectrum
        assert self.spinset.du.processed_spectrum
        assert self.spinset.ud.processed_spectrum
        assert self.spinset.uu.processed_spectrum

        self.spinset_3.process()
        assert self.spinset_3.dd.processed_spectrum
        assert self.spinset_3.du.processed_spectrum
        assert self.spinset_3.ud is None
        assert self.spinset_3.uu.processed_spectrum

        self.spinset_2.process()
        assert self.spinset_2.dd.processed_spectrum
        assert self.spinset_2.du is None
        assert self.spinset_2.ud is None
        assert self.spinset_2.uu.processed_spectrum

    def test_processing_different_reduction_options(self):
        # Check every combination of spin channel and reduction option that
        # determines the resulting wavelength axis to make sure errors
        # are raised appropriately

        standard_opts = dict(
            lo_wavelength=2.5,
            hi_wavelength=12.5,
            rebin_percent=3,
        )
        for spin_set in [self.spinset, self.spinset_3, self.spinset_2]:
            for sc in ["dd", "du", "ud", "uu"]:
                if spin_set.channels[sc] is None:
                    continue
                else:
                    for option in [
                        "lo_wavelength",
                        "hi_wavelength",
                        "rebin_percent",
                    ]:
                        spin_set.sc_opts[sc].update({option: 5})
                        with pytest.raises(ValueError):
                            spin_set.process()
                        # Reset dd_opts
                        spin_set.sc_opts[sc].update(standard_opts)


class TestPlatypusNexus(object):
    @pytest.mark.usefixtures("no_data_directory")
    @pytest.fixture(autouse=True)
    def setup_method(self, tmpdir, data_directory):
        self.pth = pjoin(data_directory, "reduce")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            self.f113 = PlatypusNexus(pjoin(self.pth, "PLP0011613.nx.hdf"))
            self.f641 = PlatypusNexus(pjoin(self.pth, "PLP0011641.nx.hdf"))

            # These PNR datasets all have different flipper settings
            self.f8861 = PlatypusNexus(
                pjoin(self.pth, "PNR_files/PLP0008861.nx.hdf")
            )
            self.f8862 = PlatypusNexus(
                pjoin(self.pth, "PNR_files/PLP0008862.nx.hdf")
            )
            self.f8863 = PlatypusNexus(
                pjoin(self.pth, "PNR_files/PLP0008863.nx.hdf")
            )
            self.f8864 = PlatypusNexus(
                pjoin(self.pth, "PNR_files/PLP0008864.nx.hdf")
            )

        self.cwd = os.getcwd()

        self.tmpdir = tmpdir.strpath
        os.chdir(self.tmpdir)
        return 0

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_chod(self):
        flight_length = self.f113.chod()
        assert_almost_equal(flight_length[0], 7141.413818359375)
        assert_almost_equal(flight_length[1], 808)

        flight_length = self.f641.chod(omega=1.8, twotheta=3.6)
        assert_almost_equal(flight_length[0], 7146.3567785516016)
        assert_almost_equal(flight_length[1], 808)

    def test_phase_angle(self):
        # TODO. Add more tests where the phase_angle isn't zero.
        phase_angle, master_opening = self.f641.phase_angle()
        assert_almost_equal(phase_angle, 0)
        assert_almost_equal(master_opening, 60)

        assert self.f641.cat.t_offset is None

    def test_background_subtract_line(self):
        # checked each step of the background subtraction with IGOR
        # so this test background correction should be correct.

        # create some test data
        xvals = np.linspace(-10, 10, 201)
        yvals = np.ceil(gauss(xvals, 0, 100, 0, 1) + 2 * xvals + 30)

        # add some reproducible random noise
        np.random.seed(1)
        yvals += np.sqrt(yvals) * np.random.randn(yvals.size)
        yvals_sd = np.sqrt(yvals)

        mask = np.zeros(201, bool)
        mask[30:70] = True
        mask[130:160] = True

        profile, profile_sd = plp.background_subtract_line(
            yvals, yvals_sd, mask
        )

        verified_data = np.load(pjoin(self.pth, "background_subtract.npy"))

        assert_almost_equal(verified_data, np.c_[profile, profile_sd])

    def test_find_specular_ridge(self):
        xvals = np.linspace(-10, 10, 201)
        yvals = np.ceil(gauss(xvals, 0, 1000, 0, 1))
        detector = np.repeat(yvals[:, np.newaxis], 1000, axis=1).T
        detector_sd = np.sqrt(detector)
        output = plp.find_specular_ridge(
            detector[np.newaxis, :], detector_sd[np.newaxis, :]
        )
        assert_(len(output) == 5)
        assert_almost_equal(output[0][0], 100)

    def test_pickle(self):
        # can we pickle a PlatypusNexus object?
        self.f8864.process()
        pkl = pickle.dumps(self.f8864)
        o = pickle.loads(pkl)
        ops, f8864 = o.processed_spectrum, self.f8864.processed_spectrum
        assert_allclose(ops["m_spec"], f8864["m_spec"])

    def test_background_subtract(self):
        # create some test data
        xvals = np.linspace(-10, 10, 201)
        yvals = np.ceil(gauss(xvals, 0, 100, 0, 1) + 2 * xvals + 30)

        # add some reproducible random noise
        np.random.seed(1)
        yvals += np.sqrt(yvals) * np.random.randn(yvals.size)
        yvals_sd = np.sqrt(yvals)

        # now make an (N, T, Y) detector image
        n_tbins = 10
        detector = np.repeat(yvals, n_tbins).reshape(xvals.size, n_tbins).T
        detector_sd = (
            np.repeat(yvals_sd, n_tbins).reshape(xvals.size, n_tbins).T
        )
        detector = detector.reshape(1, n_tbins, xvals.size)
        detector_sd = detector_sd.reshape(1, n_tbins, xvals.size)

        mask = np.zeros((1, n_tbins, 201), bool)
        mask[:, :, 30:70] = True
        mask[:, :, 130:160] = True

        det_bkg, detSD_bkg = plp.background_subtract(
            detector, detector_sd, mask
        )

        # each of the (N, T) entries should have the same background subtracted
        # entries
        verified_data = np.load(pjoin(self.pth, "background_subtract.npy"))

        it = np.nditer(detector, flags=["multi_index"])
        it.remove_axis(2)
        while not it.finished:
            profile = det_bkg[it.multi_index]
            profile_sd = detSD_bkg[it.multi_index]
            assert_almost_equal(verified_data, np.c_[profile, profile_sd])
            it.iternext()

    def test_calculate_bins(self):
        bins = plp.calculate_wavelength_bins(2.0, 18, 2.0)
        assert_almost_equal(bins[0], 1.98)
        assert_almost_equal(bins[-1], 18.18)

    def test_event(self):
        # When you use event mode processing, make sure the right amount of
        # spectra are created
        out = self.f641.process(eventmode=[0, 900, 1800], integrate=0)
        assert_(np.size(out[1], axis=0) == 2)

    def test_event_folder(self):
        self.f641.process(
            eventmode=[0, 900, 1800], integrate=0, event_folder=self.pth
        )

    def test_multiple_acquisitions(self):
        """
        TODO: add a dataset which has multiple spectra in it, and make sure it
        processes.
        """
        pass

    def test_reduction_runs(self):
        # just check it runs
        self.f641.process()

        # check that event mode reduction gives the same output as non-event
        # mode reduction.
        spectrum0 = self.f641.process(direct=False)
        assert "reduction_options" in self.f641.processed_spectrum

        spectrum1 = self.f641.process(direct=False, eventmode=[], integrate=0)
        assert_allclose(spectrum0[1][0], spectrum1[1][0], rtol=0.001)

        # check that the wavelength resolution is roughly right, between 7 and
        # 8%.
        res = (
            self.f641.processed_spectrum["m_lambda_fwhm"][0]
            / self.f641.processed_spectrum["m_lambda"][0]
        )
        assert_array_less(res, np.ones_like(res) * 0.08)
        assert_array_less(np.ones_like(res) * 0.07, res)

    def test_save_spectrum(self):
        # test saving spectrum
        self.f113.process()

        # can save the spectra by supplying a filename
        self.f113.write_spectrum_xml(pjoin(self.tmpdir, "test.xml"))
        self.f113.write_spectrum_dat(pjoin(self.tmpdir, "test.dat"))

        # can save by supplying file handle:
        with open(pjoin(self.tmpdir, "test.xml"), "wb") as f:
            self.f113.write_spectrum_xml(f)

    def test_accumulate_files(self):
        fnames = ["PLP0000708.nx.hdf", "PLP0000709.nx.hdf"]
        pths = [pjoin(self.pth, fname) for fname in fnames]
        plp.accumulate_HDF_files(pths)
        f8, f9, fadd = None, None, None

        try:
            f8 = h5py.File(pjoin(self.pth, "PLP0000708.nx.hdf"), "r")
            f9 = h5py.File(pjoin(self.pth, "PLP0000709.nx.hdf"), "r")
            fadd = h5py.File(pjoin(self.tmpdir, "ADD_PLP0000708.nx.hdf"), "r")

            f8d = f8["entry1/data/hmm"][0]
            f9d = f9["entry1/data/hmm"][0]
            faddd = fadd["entry1/data/hmm"][0]
            assert_equal(faddd, f8d + f9d)
        finally:
            if f8 is not None:
                f8.close()
            if f9 is not None:
                f9.close()
            if fadd is not None:
                fadd.close()

    def test_accumulate_files_reduce(self):
        # test by adding a file to itself. Should have smaller stats
        fnames = ["PLP0000708.nx.hdf", "PLP0000708.nx.hdf"]
        pths = [pjoin(self.pth, fname) for fname in fnames]
        plp.accumulate_HDF_files(pths)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            # it should be processable
            fadd = PlatypusNexus(pjoin(os.getcwd(), "ADD_PLP0000708.nx.hdf"))
            fadd.process()

            # it should also be reduceable
            reducer = PlatypusReduce(pjoin(self.pth, "PLP0000711.nx.hdf"))

            datasets, reduced = reducer.reduce(
                pjoin(os.getcwd(), "ADD_PLP0000708.nx.hdf")
            )
            assert_("y" in reduced)

            # the error bars should be smaller
            datasets2, reduced2 = reducer.reduce(
                pjoin(self.pth, "PLP0000708.nx.hdf")
            )

            assert_(np.all(reduced["y_err"] < reduced2["y_err"]))

    def test_manual_beam_find(self):
        # you can specify a function that finds where the specular ridge is.
        def manual_beam_find(detector, detector_sd, name):
            beam_centre = np.zeros(len(detector))
            beam_sd = np.zeros(len(detector))
            beam_centre += 50
            beam_sd += 5
            return beam_centre, beam_sd, np.array([40]), np.array([60]), [[]]

        # the manual beam find is only mandatory when peak_pos == -1.
        # the beam_sd is much larger than the beam divergence, so a warning
        # should be raised.
        with pytest.warns(UserWarning):
            self.f113.process(manual_beam_find=manual_beam_find, peak_pos=-1)
        assert_equal(self.f113.processed_spectrum["m_beampos"][0], 50)

        # manual beam finding also specifies the lower and upper pixel of the
        # foreground
        assert_equal(self.f113.processed_spectrum["lopx"][0], 40)
        assert_equal(self.f113.processed_spectrum["hipx"][0], 60)

    def test_fore_back_region(self):
        # calculation of foreground and background regions is done correctly
        centres = np.array([100.0, 90.0])
        sd = np.array([5.0, 11.5])
        lopx, hipx, background_pixels = fore_back_region(centres, sd)

        assert_(len(lopx) == 2)
        assert_(len(hipx) == 2)
        assert_(len(background_pixels) == 2)
        assert_(isinstance(lopx[0], numbers.Integral))
        assert_(isinstance(hipx[0], numbers.Integral))

        calc_lower = np.floor(centres - sd * EXTENT_MULT)
        assert_equal(lopx, calc_lower)
        calc_higher = np.ceil(centres + sd * EXTENT_MULT)
        assert_equal(hipx, calc_higher)

        y1 = np.atleast_1d(np.round(lopx - PIXEL_OFFSET).astype("int"))
        y0 = np.atleast_1d(
            np.round(lopx - PIXEL_OFFSET - EXTENT_MULT * sd).astype("int")
        )

        y2 = np.atleast_1d(np.round(hipx + PIXEL_OFFSET).astype("int"))
        y3 = np.atleast_1d(
            np.round(hipx + PIXEL_OFFSET + EXTENT_MULT * sd).astype("int")
        )

        bp = np.r_[np.arange(y0[0], y1[0] + 1), np.arange(y2[0], y3[0] + 1)]

        assert_equal(bp, background_pixels[0])

    def test_basename_datafile(self):
        # check that the right basename is returned
        pth = "a/b/c.nx.hdf"
        assert_(basename_datafile(pth) == "c")

        pth = "c.nx.hdf"
        assert_(basename_datafile(pth) == "c")

    def test_floodfield_correction(self):
        # check that flood field calculation works
        # the values were worked out by hand on a randomly
        # generated array
        test_norm = np.array(
            [1.12290503, 1.23743017, 0.8603352, 0.70111732, 1.07821229]
        )
        test_norm_sd = np.array(
            [0.05600541, 0.05879208, 0.04902215, 0.04425413, 0.05487956]
        )

        fname = os.path.join(self.pth, "flood.h5")
        with h5py.File(fname, "r") as f:
            norm, norm_sd = create_detector_norm(f, 3.5, -3.5, axis=3)

            assert_almost_equal(norm, test_norm, 6)
            assert_almost_equal(norm_sd, test_norm_sd, 6)

            norm, norm_sd = create_detector_norm(f, -3.5, 3.5, axis=3)
            assert_almost_equal(norm, test_norm, 6)
            assert_almost_equal(norm_sd, test_norm_sd, 6)

    def test_PNR_metadata(self):
        self.f8861.process()
        self.f8862.process()
        self.f8863.process()
        self.f8864.process()

        # Check that we can read all of the flipper statuses in files

        # Flipper 1 on, flipper 2 on
        assert_almost_equal(self.f8861.cat.cat["pol_flip_current"], 5.0)
        assert_almost_equal(self.f8861.cat.cat["anal_flip_current"], 4.5)

        # Flipper 1 on, flipper 2 off
        assert_almost_equal(self.f8862.cat.cat["anal_flip_current"], 0)
        assert_almost_equal(self.f8862.cat.cat["pol_flip_current"], 5.0)

        # Flipper 1 off, flipper 2 on
        assert_almost_equal(self.f8863.cat.cat["anal_flip_current"], 4.5)
        assert_almost_equal(self.f8863.cat.cat["pol_flip_current"], 0)

        # Flipper 1 off, flipper 2 off
        assert_almost_equal(self.f8864.cat.cat["anal_flip_current"], 0)
        assert_almost_equal(self.f8864.cat.cat["pol_flip_current"], 0)

        # Check SpinChannel for each file
        assert self.f8861.spin_state == SpinChannel.UP_UP
        assert self.f8862.spin_state == SpinChannel.UP_DOWN
        assert self.f8863.spin_state == SpinChannel.DOWN_UP
        assert self.f8864.spin_state == SpinChannel.DOWN_DOWN

        # test spin channel setting
        # non spin analysed. mode is POL, not POLANAL
        pn = PlatypusNexus(pjoin(self.pth, "PLP0016427.nx.hdf"))
        assert pn.spin_state == SpinChannel.UP_UP
        pn = PlatypusNexus(pjoin(self.pth, "PLP0016426.nx.hdf"))
        assert pn.spin_state == SpinChannel.DOWN_DOWN

    def test_PNR_magnet_read(self):
        self.f8861.process()
        # Check magnetic field sensors
        assert_almost_equal(self.f8861.cat.cat["magnet_current_set"], 0)
        assert_almost_equal(self.f8861.cat.cat["magnet_output_current"], 0.001)


class TestSpatzNexus:
    @pytest.mark.usefixtures("no_data_directory")
    @pytest.fixture(autouse=True)
    def setup_method(self, tmpdir, data_directory):
        self.pth = pjoin(data_directory, "reduce")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            self.f342 = SpatzNexus(pjoin(self.pth, "SPZ0000342.nx.hdf"))
        self.cwd = os.getcwd()

        self.tmpdir = tmpdir.strpath
        os.chdir(self.tmpdir)
        return 0

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_chod(self):
        flight_length = self.f342.chod()
        assert_almost_equal(flight_length[0], 8062.0232)
        assert_almost_equal(flight_length[1], 479.9536, decimal=4)

    def test_phase_angle(self):
        assert_allclose(self.f342.cat.master_phase_offset, -25.90)
        assert self.f342.cat.master == 1
        assert self.f342.cat.slave == 2
        assert self.f342.cat.t_offset is None
        assert_allclose(self.f342.cat.frequency, 25)
        assert_allclose(self.f342.cat.phase, 34.22)
        assert_allclose(self.f342.cat.poff_c2_slave_1_master[0], -0.22)

        phase_angle, master_opening = self.f342.phase_angle()
        # chopper 1 has an opening of 26 degrees
        assert_almost_equal(master_opening, 26)
        assert_allclose(phase_angle, 0, atol=1e-5)

        toff = self.f342.time_offset(-25.90, 26, 25, 0.0, None, None, None)
        assert_allclose(toff, 5.5555555555555)

        toff = self.f342.time_offset(
            -25.90,
            26,
            25,
            0.0,
            None,
            None,
            None,
            t_offset=1438.8888888888888,
        )
        assert_allclose(toff, 0, atol=1e-12)


@pytest.mark.usefixtures("no_data_directory")
def test_catalogue(data_directory):
    pth = pjoin(data_directory, "reduce")
    catalogue(0, 10000000, data_folder=pth, prefix="PLP")
    catalogue(0, 10000000, data_folder=pth, prefix="SPZ")


@pytest.mark.usefixtures("no_data_directory")
def test_Catalogue_pickle(data_directory):
    pth = pjoin(data_directory, "reduce")
    f113 = pjoin(pth, "PLP0011613.nx.hdf")
    with h5py.File(f113) as f:
        c = PlatypusCatalogue(f)

    assert c.datafile_number == 11613
    assert c.prefix == "PLP"
    pkl = pickle.dumps(c)
    pickle.loads(pkl)



"""
*Calculates the specular (Neutron or X-ray) reflectivity from a stratified
series of layers.

The refnx code is distributed under the following license:

Copyright (c) 2015 A. R. J. Nelson, ANSTO

Permission to use and redistribute the source code or binary forms of this
software and its documentation, with or without modification is hereby
granted provided that the above notice of copyright, these terms of use,
and the disclaimer of warranty below appear in the source code and
documentation, and that none of the names of above institutions or
authors appear in advertising or endorsement of works derived from this
software without specific prior written permission from all parties.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THIS SOFTWARE.

"""
from functools import reduce

import jax.numpy as jnp
from jax import jit
from jax.ops import index, index_add, index_update

TINY = 1e-30


def jabeles(q, layers, scale=1.0, bkg=0, threads=0):
    qvals = q.astype(jnp.float64)
    flatq = qvals.ravel()

    nlayers = layers.shape[0] - 2
    npnts = flatq.size

    mi00 = jnp.ones((npnts, nlayers + 1), jnp.complex128)

    sld = jnp.zeros(nlayers + 2, jnp.complex128)

    # addition of TINY is to ensure the correct branch cut
    # in the complex sqrt calculation of kn.
    sld = index_add(
        sld,
        index[1:],
        ((layers[1:, 1] - layers[0, 1]) + 1j * (jnp.abs(layers[1:, 2]) + TINY))
        * 1.0e-6,
    )
    kn = jnp.sqrt(flatq[:, jnp.newaxis] ** 2.0 / 4.0 - 4.0 * jnp.pi * sld)
    # reflectances for each layer
    # rj.shape = (npnts, nlayers + 1)
    damping = jnp.exp(-2.0 * kn[:, :-1] * kn[:, 1:] * layers[1:, 3] ** 2)
    rj = (kn[:, :-1] - kn[:, 1:]) / (kn[:, :-1] + kn[:, 1:]) * damping

    # characteristic matrices for each layer
    # miNN.shape = (npnts, nlayers + 1)
    if nlayers:
        mi00 = index_update(
            mi00,
            index[:, 1:],
            jnp.exp(kn[:, 1:-1] * 1j * jnp.fabs(layers[1:-1, 0])),
        )
    mi11 = 1.0 / mi00
    mi10 = rj * mi11
    mi01 = rj * mi00

    mi = jnp.zeros((npnts, nlayers + 1, 2, 2), jnp.complex128)

    mi = index_update(
        mi,
        index[:, :, 0, 0],
        mi00,
    )
    mi = index_update(
        mi,
        index[:, :, 0, 1],
        mi01,
    )
    mi = index_update(
        mi,
        index[:, :, 1, 1],
        mi11,
    )
    mi = index_update(
        mi,
        index[:, :, 1, 0],
        mi10,
    )

    sub = [jnp.squeeze(v) for v in jnp.hsplit(mi, nlayers + 1)]
    mrtot = reduce(jnp.matmul, sub[1:], sub[0])

    r = mrtot[:, 1, 0] / mrtot[:, 0, 0]
    reflectivity = r * jnp.conj(r) * scale
    reflectivity = index_add(reflectivity, ..., bkg)

    return jnp.real(jnp.reshape(reflectivity, qvals.shape))


# abeles_jax = jabeles
abeles_jax = jit(jabeles)



"""
refnx is distributed under the following license:

Copyright (c) 2015 A. R. J. Nelson, ANSTO

Permission to use and redistribute the source code or binary forms of this
software and its documentation, with or without modification is hereby
granted provided that the above notice of copyright, these terms of use,
and the disclaimer of warranty below appear in the source code and
documentation, and that none of the names of above institutions or
authors appear in advertising or endorsement of works derived from this
software without specific prior written permission from all parties.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THIS SOFTWARE.

"""
import time
import datetime
import pickle
import warnings
import glob
import os

import numpy as np

import ipywidgets as widgets
import traitlets
from traitlets import HasTraits
import matplotlib.gridspec as gridspec

from refnx.reflect import Slab, ReflectModel
from refnx.dataset import ReflectDataset
from refnx.analysis import Objective, CurveFitter, Transform
from refnx._lib import flatten, possibly_open_file


class ReflectModelView(HasTraits):
    """
    An ipywidgets viewport of a `refnx.reflect.ReflectModel`.

    Parameters
    ----------
    reflect_model: refnx.reflect.ReflectModel

    Notes
    -----
    Use the `model_box` property to view/modify the ReflectModel parameters.
    Use the `limits_box` property to view the limits for the varying
    parameters.
    Observe the `view_changed` traitlet to determine when widget values are
    changed.
    Observe the `view_redraw` traitlet to determine when a complete redraw
    of the view is required (because the number of widgets has changed for
    example).

    """

    # traitlet to say when params were last altered
    view_changed = traitlets.Float(time.time())

    # traitlet to ask when a redraw of the GUI is requested.
    # e.g. the number of layers has changed, or there are a
    # different number of fitted parameters requiring
    # different limit widgets.
    view_redraw = traitlets.Float(time.time())

    # number of varying parameters in the model
    num_varying = traitlets.Int(0)

    def __init__(self, reflect_model):
        super().__init__()

        self.model = reflect_model
        self.structure_view = StructureView(self.model.structure)
        self.last_selected_param = None
        self.param_widgets_link = {}

        slab_views = self.structure_view.slab_views
        slab_views[0].w_thick.disabled = True
        slab_views[0].c_thick.disabled = True
        slab_views[0].w_rough.disabled = True
        slab_views[0].c_rough.disabled = True
        slab_views[-1].w_thick.disabled = True
        slab_views[-1].c_thick.disabled = True

        # got to listen to all the slab views
        for slab_view in slab_views:
            slab_view.observe(
                self._on_slab_params_modified, names=["view_changed"]
            )

        # if you'd like to change the number of layers
        self.w_layers = widgets.BoundedIntText(
            description="Number of layers",
            value=len(slab_views) - 2,
            min=0,
            max=1000,
            style={"description_width": "120px"},
            continuous_update=False,
        )

        self.w_layers.observe(self._on_change_layers, names=["value"])

        # where you're going to add/remove layers
        # varying layers is a flag to say if you're currently in the process
        # of adding/removing layers
        self._varying_layers = False
        self._location = None
        self.ok_button = None
        self.cancel_button = None

        # associated with ReflectModel
        p = reflect_model.scale
        self.w_scale = widgets.FloatText(
            value=p.value,
            description="scale",
            step=0.01,
            style={"description_width": "120px"},
        )
        self.c_scale = widgets.Checkbox(value=p.vary)
        self.scale_low_limit = widgets.FloatText(value=p.bounds.lb, step=0.01)
        self.scale_hi_limit = widgets.FloatText(value=p.bounds.ub, step=0.01)

        p = reflect_model.bkg
        self.w_bkg = widgets.FloatText(
            value=p.value,
            description="background",
            step=1e-7,
            style={"description_width": "120px"},
        )
        self.c_bkg = widgets.Checkbox(value=reflect_model.bkg.vary)
        self.bkg_low_limit = widgets.FloatText(p.bounds.lb, step=1e-8)
        self.bkg_hi_limit = widgets.FloatText(value=p.bounds.ub, step=1e-7)

        p = reflect_model.dq
        self.w_dq = widgets.BoundedFloatText(
            value=p.value, description="dq/q", step=0.1, min=0, max=20.0
        )
        self.c_dq = widgets.Checkbox(value=reflect_model.dq.vary)
        self.dq_low_limit = widgets.BoundedFloatText(
            value=p.bounds.lb, min=0, max=20, step=0.1
        )
        self.dq_hi_limit = widgets.BoundedFloatText(
            value=p.bounds.ub, min=0, max=20, step=0.1
        )

        self.c_scale.style.description_width = "0px"
        self.c_bkg.style.description_width = "0px"
        self.c_dq.style.description_width = "0px"
        self.do_fit_button = widgets.Button(description="Do Fit")
        self.to_code_button = widgets.Button(description="To code")
        self.save_model_button = widgets.Button(description="Save Model")
        self.load_model_button = widgets.Button(description="Load Model")

        widget_list = [
            self.w_scale,
            self.c_scale,
            self.w_bkg,
            self.c_bkg,
            self.w_dq,
            self.c_dq,
        ]

        limits_widgets_list = [
            self.scale_low_limit,
            self.scale_hi_limit,
            self.bkg_low_limit,
            self.bkg_hi_limit,
            self.dq_low_limit,
            self.dq_hi_limit,
        ]

        for widget in widget_list:
            widget.observe(self._on_model_params_modified, names=["value"])

        for widget in limits_widgets_list:
            widget.observe(self._on_model_limits_modified, names=["value"])

        # button to create default limits
        self.default_limits_button = widgets.Button(
            description="Set default limits"
        )
        self.default_limits_button.on_click(self.default_limits)

        # widgets for easy model change
        self.model_slider = widgets.FloatSlider()
        self.model_slider.layout = widgets.Layout(width="100%")
        self.model_slider_link = None
        self.model_slider_min = widgets.FloatText()
        self.model_slider_min.layout = widgets.Layout(width="10%")
        self.model_slider_max = widgets.FloatText()
        self.model_slider_max.layout = widgets.Layout(width="10%")
        self.model_slider_max.observe(
            self._on_slider_limits_modified, names=["value"]
        )
        self.model_slider_min.observe(
            self._on_slider_limits_modified, names=["value"]
        )
        self.last_selected_param = None

        self.num_varying = len(self.model.parameters.varying_parameters())

        self._link_param_widgets()

    def _on_model_params_modified(self, change):
        """
        Called when ReflectModel parameters are varied.
        """
        d = self.param_widgets_link

        for par in [self.model.scale, self.model.bkg, self.model.dq]:
            idx = id(par)
            wids = d[idx]

            if change["owner"] in wids:
                loc = wids.index(change["owner"])
                if loc == 0:
                    par.value = wids[0].value

                    # this captures when the user starts modifying a different
                    # parameter
                    self._possibly_link_slider(change["owner"])

                    self.view_changed = time.time()
                    break
                elif loc == 1:
                    # you are changing the number of varying parameters
                    par.vary = wids[1].value

                    # need to rebuild the limit widgets, achieved by redrawing
                    # box
                    # set the number of varying parameters
                    self.num_varying = len(
                        self.model.parameters.varying_parameters()
                    )

                    self.view_redraw = time.time()
                    break
                else:
                    return

        # this captures when the user starts modifying a different parameter
        self._possibly_link_slider(change["owner"])

    def _on_model_limits_modified(self, change):
        """
        When a limit widget is changed, update corresponding limits in the
        underlying ReflectModel.
        """
        d = self.param_widgets_link
        for par in [self.model.scale, self.model.bkg, self.model.dq]:
            idx = id(par)
            wids = d[idx]

            if change["owner"] in wids:
                loc = wids.index(change["owner"])
                if loc == 2:
                    par.bounds.lb = wids[2].value
                    break
                elif loc == 3:
                    par.bounds.ub = wids[3].value
                    break

    def default_limits(self, change):
        """
        Makes default limits for the parameters being varied
        """
        varying_parameters = self.model.parameters.varying_parameters()

        for par in varying_parameters:
            par.bounds.lb = min(0, 2 * par.value)
            par.bounds.ub = max(0, 2 * par.value)

        self.refresh()

    def _on_slab_params_modified(self, change):
        """
        Called when slab parameters are varied.
        """
        # this captures when the user starts modifying a different parameter
        if isinstance(change["owner"].param_being_varied, widgets.Checkbox):
            # you are changing the number of fitted parameters

            # set the number of varying parameters
            self.num_varying = len(self.model.parameters.varying_parameters())

            # need to rebuild the limit widgets, achieved by redrawing box
            self.view_redraw = time.time()
        else:
            self._possibly_link_slider(change["owner"].param_being_varied)
            self.view_changed = time.time()

    def _possibly_link_slider(self, change_owner):
        """
        When a ReflectModel value is changed link a slider widget to the
        parameter that is being varied.
        """
        if change_owner is not self.last_selected_param:
            self.last_selected_param = change_owner
            if self.model_slider_link is not None:
                self.model_slider_link.unlink()
            self.model_slider_link = widgets.link(
                (self.last_selected_param, "value"),
                (self.model_slider, "value"),
            )
            self.model_slider_max.value = max(
                0, 2.0 * self.last_selected_param.value
            )
            self.model_slider_min.value = min(
                0, 2.0 * self.last_selected_param.value
            )

    def _on_slider_limits_modified(self, change):
        """
        Callback when adjusting the min, max widgets for the main slider
        widget.
        """
        self.model_slider.max = self.model_slider_max.value
        self.model_slider.min = self.model_slider_min.value
        self.model_slider.step = (
            self.model_slider.max - self.model_slider.min
        ) / 200.0

    def _on_change_layers(self, change):
        self.ok_button = widgets.Button(description="OK")
        if change["new"] > change["old"]:
            description = "Insert before which layer?"
            min_loc = 1
            max_loc = len(self.model.structure) - 2 + 1
            self.ok_button.on_click(self._increase_layers)
        elif change["new"] < change["old"]:
            min_loc = 1
            max_loc = (
                len(self.model.structure)
                - 2
                - (change["old"] - change["new"])
                + 1
            )
            description = "Remove from which layer?"
            self.ok_button.on_click(self._decrease_layers)
        else:
            return
        self._varying_layers = True
        self.w_layers.disabled = True
        self.do_fit_button.disabled = True
        self.to_code_button.disabled = True
        self.save_model_button.disabled = True
        self.load_model_button.disabled = True

        self._location = widgets.BoundedIntText(
            value=min_loc,
            description=description,
            min=min_loc,
            max=max_loc,
            style={"description_width": "initial"},
        )
        self.cancel_button = widgets.Button(description="Cancel")
        self.cancel_button.on_click(self._cancel_layers)
        self.view_redraw = time.time()

    def _increase_layers(self, b):
        self.w_layers.disabled = False
        self.do_fit_button.disabled = False
        self.to_code_button.disabled = False
        self.save_model_button.disabled = False
        self.load_model_button.disabled = False

        how_many = self.w_layers.value - (len(self.model.structure) - 2)
        loc = self._location.value

        for i in range(how_many):
            slab = Slab(0, 0, 3)
            slab.thick.bounds = (0, 2 * slab.thick.value)
            slab.sld.real.bounds = (0, 2 * slab.sld.real.value)
            slab.sld.imag.bounds = (0, 2 * slab.sld.imag.value)
            slab.rough.bounds = (0, 2 * slab.rough.value)

            slab_view = SlabView(slab)
            self.model.structure.insert(loc, slab)
            self.structure_view.slab_views.insert(loc, slab_view)
            slab_view.observe(self._on_slab_params_modified)

        rename_params(self.model.structure)
        self._varying_layers = False

        # set the number of varying parameters
        self.num_varying = len(self.model.parameters.varying_parameters())

        self.view_redraw = time.time()

    def _decrease_layers(self, b):
        self.w_layers.disabled = False
        self.do_fit_button.disabled = False
        self.to_code_button.disabled = False
        self.save_model_button.disabled = False
        self.load_model_button.disabled = False

        loc = self._location.value
        how_many = len(self.model.structure) - 2 - self.w_layers.value
        for i in range(how_many):
            self.model.structure.pop(loc)
            slab_view = self.structure_view.slab_views.pop(loc)
            slab_view.unobserve_all()

        rename_params(self.model.structure)
        self._varying_layers = False

        # set the number of varying parameters
        self.num_varying = len(self.model.parameters.varying_parameters())

        self.view_redraw = time.time()

    def _link_param_widgets(self):
        """
        Creates a dictionary of {parameter: (associated_widgets_tuple)}.
        """

        # link parameters to widgets (value, checkbox,
        #                             upperlim, lowerlim)
        self.param_widgets_link = {}
        d = self.param_widgets_link
        model = self.model

        d[id(model.scale)] = (
            self.w_scale,
            self.c_scale,
            self.scale_low_limit,
            self.scale_hi_limit,
        )
        d[id(model.bkg)] = (
            self.w_bkg,
            self.c_bkg,
            self.bkg_low_limit,
            self.bkg_hi_limit,
        )
        d[id(model.dq)] = (
            self.w_dq,
            self.c_dq,
            self.dq_low_limit,
            self.dq_hi_limit,
        )

    def _cancel_layers(self, b):
        # disable the change layers widget to prevent recursion
        self.w_layers.unobserve(self._on_change_layers, names="value")
        self.w_layers.value = len(self.model.structure) - 2
        self.w_layers.observe(self._on_change_layers, names="value")
        self.w_layers.disabled = False
        self.do_fit_button.disabled = False
        self.to_code_button.disabled = False
        self.save_model_button.disabled = False
        self.load_model_button.disabled = False

        self._varying_layers = False
        self.view_redraw = time.time()

    def refresh(self):
        """
        Updates the widget values from the underlying `ReflectModel`.
        """
        for par in [self.model.scale, self.model.bkg, self.model.dq]:
            wid = self.param_widgets_link[id(par)]
            wid[0].value = par.value
            wid[1].value = par.vary
            wid[2].value = par.bounds.lb
            wid[3].value = par.bounds.ub

        slab_views = self.structure_view.slab_views

        for slab_view in slab_views:
            slab_view.refresh()

    @property
    def model_box(self):
        """
        `ipywidgets.Vbox` displaying model relevant widgets.
        """
        output = [
            self.w_layers,
            widgets.HBox([self.w_scale, self.c_scale, self.w_dq, self.c_dq]),
            widgets.HBox([self.w_bkg, self.c_bkg]),
            self.structure_view.box,
            widgets.HBox(
                [
                    self.model_slider_min,
                    self.model_slider,
                    self.model_slider_max,
                ]
            ),
        ]

        if self._varying_layers:
            output.append(
                widgets.HBox(
                    [self._location, self.ok_button, self.cancel_button]
                )
            )

        output.append(
            widgets.HBox(
                [
                    self.do_fit_button,
                    self.to_code_button,
                    self.save_model_button,
                    self.load_model_button,
                ]
            )
        )

        return widgets.VBox(output)

    @property
    def limits_box(self):
        varying_pars = self.model.parameters.varying_parameters()
        hboxes = [self.default_limits_button]

        d = {}
        d.update(self.param_widgets_link)

        slab_views = self.structure_view.slab_views
        for slab_view in slab_views:
            d.update(slab_view.param_widgets_link)

        for par in varying_pars:
            name = widgets.Text(par.name)
            name.disabled = True

            val, check, ll, ul = d[id(par)]

            hbox = widgets.HBox([name, ll, val, ul])
            hboxes.append(hbox)

        return widgets.VBox(hboxes)


class StructureView:
    def __init__(self, structure):
        self.structure = structure
        self.slab_views = [SlabView(slab) for slab in structure]

    @property
    def box(self):
        layout = widgets.Layout(flex="1 1 auto", width="auto")
        label_row = widgets.HBox(
            [
                widgets.HTML("thick", layout=layout),
                widgets.HTML("sld", layout=layout),
                widgets.HTML("isld", layout=layout),
                widgets.HTML("rough", layout=layout),
            ]
        )

        hboxes = [label_row]
        hboxes.extend([view.box for view in self.slab_views])
        # add in layer numbers
        self.slab_views[0].w_thick.description = "fronting"
        self.slab_views[-1].w_thick.description = "backing"
        for i in range(1, len(self.slab_views) - 1):
            self.slab_views[i].w_thick.description = str(i)

        return widgets.VBox(hboxes)


class SlabView(HasTraits):
    """
    An ipywidgets viewport of a `refnx.reflect.Slab`.

    Parameters
    ----------
    slab: refnx.reflect.Slab

    Notes
    -----
    An ipywidgets viewport of a `refnx.reflect.Slab`.
    Use the `box` property to view/modify the `Slab` parameters.
    Observe the `view_changed` traitlet to determine when widget values are
    changed.
    """

    # traitlet to say when params were last altered
    view_changed = traitlets.Float(time.time())

    def __init__(self, slab):
        self.slab = slab
        self.param_widgets_link = {}
        self.widgets_param_link = {}

        self.param_being_varied = None

        p = slab.thick
        self.w_thick = widgets.FloatText(value=p.value, step=1)
        self.widgets_param_link[self.w_thick] = p
        self.c_thick = widgets.Checkbox(value=p.vary)
        self.thick_low_limit = widgets.FloatText(value=p.bounds.lb, step=1)
        self.thick_hi_limit = widgets.FloatText(value=p.bounds.ub, step=1)

        p = slab.sld.real
        self.w_sld = widgets.FloatText(value=p.value, step=0.01)
        self.widgets_param_link[self.w_sld] = p
        self.c_sld = widgets.Checkbox(value=p.vary)
        self.sld_low_limit = widgets.FloatText(value=p.bounds.lb, step=0.01)
        self.sld_hi_limit = widgets.FloatText(value=p.bounds.ub, step=0.01)
        p = slab.sld.imag
        self.w_isld = widgets.FloatText(value=p.value, step=0.01)
        self.widgets_param_link[self.w_isld] = p
        self.c_isld = widgets.Checkbox(value=p.vary)
        self.isld_low_limit = widgets.FloatText(value=p.bounds.lb, step=0.01)
        self.isld_hi_limit = widgets.FloatText(value=p.bounds.ub, step=0.01)

        p = slab.rough
        self.w_rough = widgets.FloatText(value=p, step=1)
        self.widgets_param_link[self.w_rough] = p
        self.c_rough = widgets.Checkbox(value=p.vary)
        self.rough_low_limit = widgets.FloatText(p.bounds.lb, step=0.01)
        self.rough_hi_limit = widgets.FloatText(value=p.bounds.ub, step=0.01)

        self._widget_list = [
            self.w_thick,
            self.c_thick,
            self.w_sld,
            self.c_sld,
            self.w_isld,
            self.c_isld,
            self.w_rough,
            self.c_rough,
        ]
        self._limits_list = [
            self.thick_low_limit,
            self.thick_hi_limit,
            self.sld_low_limit,
            self.sld_hi_limit,
            self.isld_low_limit,
            self.isld_hi_limit,
            self.rough_low_limit,
            self.rough_hi_limit,
        ]

        # link widgets to observers
        for widget in [self.w_thick, self.w_sld, self.w_isld, self.w_rough]:
            widget.style.description_width = "0px"
            widget.observe(self._on_slab_values_modified, names="value")
        self.w_thick.style.description_width = "50px"

        for widget in [self.c_thick, self.c_sld, self.c_isld, self.c_rough]:
            widget.style.description_width = "0px"
            widget.observe(self._on_slab_varies_modified, names="value")

        for widget in self._limits_list:
            widget.observe(self._on_slab_limits_modified, names="value")

        self._link_param_widgets()

    def _on_slab_values_modified(self, change):
        d = self.widgets_param_link
        d[change["owner"]].value = change["owner"].value

        self.param_being_varied = change["owner"]
        self.view_changed = time.time()

    def _on_slab_varies_modified(self, change):
        d = self.param_widgets_link
        slab = self.slab

        for par in flatten(slab.parameters):
            if id(par) in d and change["owner"] in d[id(par)]:
                wids = d[id(par)]
                par.vary = wids[1].value
                break

        self.param_being_varied = change["owner"]
        self.view_changed = time.time()

    def _on_slab_limits_modified(self, change):
        slab = self.slab
        d = self.param_widgets_link

        for par in flatten(slab.parameters):
            if id(par) in d and change["owner"] in d[id(par)]:
                wids = d[id(par)]
                loc = wids.index(change["owner"])
                if loc == 2:
                    par.bounds.lb = wids[loc].value
                    break
                elif loc == 3:
                    par.bounds.ub = wids[loc].value
                    break
                else:
                    return

    def _link_param_widgets(self):
        """
        Creates a dictionary of {parameter: (associated_widgets_tuple)}.
        """
        # link parameters to widgets (value, checkbox,
        #                             upperlim, lowerlim)
        d = self.param_widgets_link

        d[id(self.slab.thick)] = (
            self.w_thick,
            self.c_thick,
            self.thick_low_limit,
            self.thick_hi_limit,
        )
        d[id(self.slab.sld.real)] = (
            self.w_sld,
            self.c_sld,
            self.sld_low_limit,
            self.sld_hi_limit,
        )
        d[id(self.slab.sld.imag)] = (
            self.w_isld,
            self.c_isld,
            self.isld_low_limit,
            self.isld_hi_limit,
        )

        d[id(self.slab.rough)] = (
            self.w_rough,
            self.c_rough,
            self.rough_low_limit,
            self.rough_hi_limit,
        )

    def refresh(self):
        """
        Updates the widget values from the underlying `Slab` parameters.
        """
        d = self.param_widgets_link

        ids = {id(p): p for p in flatten(self.slab.parameters) if id(p) in d}
        for idx, par in ids.items():
            widgets = d[idx]
            widgets[0].value = par.value
            widgets[1].value = par.vary
            widgets[2].value = par.bounds.lb
            widgets[3].value = par.bounds.ub

    @property
    def box(self):
        return widgets.HBox(self._widget_list)


class Motofit:
    """
    An interactive slab modeller (Jupyter/ipywidgets based) for Neutron and
    X-ray reflectometry data.

    The interactive modeller is designed to be used in a Jupyter notebook.

    >>> # specify that plots are in a separate graph window
    >>> %matplotlib qt

    >>> # alternately if you want the graph to be embedded in the notebook use
    >>> # %matplotlib notebook

    >>> from refnx.reflect import Motofit
    >>> # create an instance of the modeller
    >>> app = Motofit()
    >>> # display it in the notebook by calling the object with a datafile.
    >>> app('dataset1.txt')
    >>> # lets fit a different dataset
    >>> app2 = Motofit()
    >>> app2('dataset2.txt')

    The `Motofit` instance has several useful attributes that can be used in
    other cells. For example, one can access the `objective` and `curvefitter`
    attributes for more advanced fitting functionality than is available in the
    GUI. A `code` attribute can be used to retrieve a Python code fragment that
    can be used as a basis for developing more complicated models, such as
    interparameter constraints, global fitting, etc.

    Attributes
    ----------
    dataset: :class:`refnx.dataset.Data1D`
        The dataset associated with the modeller
    model: :class:`refnx.reflect.ReflectModel`
        Calculates a theoretical model, from an interfacial structure
        (`model.Structure`).
    objective: :class:`refnx.analysis.Objective`
        The Objective that allows one to compare the model against the data.
    fig: :class:`matplotlib.figure.Figure`
        Graph displaying the data.

    """

    def __init__(self):
        # attributes for the graph
        # for the graph
        self.qmin = 0.005
        self.qmax = 0.5
        self.qpnt = 1000
        self.fig = None

        self.ax_data = None
        self.ax_residual = None
        self.ax_sld = None
        # gridspecs specify how the plots are laid out. Gridspec1 is when the
        # residuals plot is displayed. Gridspec2 is when it's not visible
        self._gridspec1 = gridspec.GridSpec(
            2, 2, height_ratios=[5, 1], width_ratios=[1, 1], hspace=0.01
        )
        self._gridspec2 = gridspec.GridSpec(1, 2)

        self.theoretical_plot = None
        self.theoretical_plot_sld = None

        # attributes for a user dataset
        self.dataset = None
        self.objective = None
        self._curvefitter = None
        self.data_plot = None
        self.residuals_plot = None
        self.data_plot_sld = None

        self.dataset_name = widgets.Text(description="dataset:")
        self.dataset_name.disabled = True
        self.chisqr = widgets.FloatText(description="chi-squared:")
        self.chisqr.disabled = True

        # fronting
        slab0 = Slab(0, 0, 0)
        slab1 = Slab(25, 3.47, 3)
        slab2 = Slab(0, 2.07, 3)

        structure = slab0 | slab1 | slab2
        rename_params(structure)
        self.model = ReflectModel(structure)
        structure = slab0 | slab1 | slab2
        self.model = ReflectModel(structure)

        # give some default parameter limits
        self.model.scale.bounds = (0.1, 2)
        self.model.bkg.bounds = (1e-8, 2e-5)
        self.model.dq.bounds = (0, 20)
        for slab in self.model.structure:
            slab.thick.bounds = (0, 2 * slab.thick.value)
            slab.sld.real.bounds = (0, 2 * slab.sld.real.value)
            slab.sld.imag.bounds = (0, 2 * slab.sld.imag.value)
            slab.rough.bounds = (0, 2 * slab.rough.value)

        # the main GUI widget
        self.display_box = widgets.VBox()

        self.tab = widgets.Tab()
        self.tab.set_title(0, "Model")
        self.tab.set_title(1, "Limits")
        self.tab.set_title(2, "Options")
        self.tab.observe(self._on_tab_changed, names="selected_index")

        # an output area for messages.
        self.output = widgets.Output()

        # options tab
        self.plot_type = widgets.Dropdown(
            options=["lin", "logY", "YX4", "YX2"],
            value="lin",
            description="Plot Type:",
            disabled=False,
        )
        self.plot_type.observe(self._on_plot_type_changed, names="value")
        self.use_weights = widgets.RadioButtons(
            options=["Yes", "No"],
            value="Yes",
            description="use dataset weights?",
            style={"description_width": "initial"},
        )
        self.use_weights.observe(self._on_use_weights_changed, names="value")
        self.transform = Transform("lin")
        self.display_residuals = widgets.Checkbox(
            value=False, description="Display residuals"
        )
        self.display_residuals.observe(
            self._on_display_residuals_changed, names="value"
        )

        self.model_view = None
        self.set_model(self.model)

    def save_model(self, *args, f=None):
        """
        Serialise a model to a pickle file.
        If `f` is not specified then the file name is constructed from the
        current dataset name; if there is no current dataset then the filename
        is constructed from the current time. These constructed filenames will
        be in the current working directory, for a specific save location `f`
        must be provided.
        This method is only intended to be used to serialise models created by
        this interactive Jupyter widget modeller.

        Parameters
        ----------
        f: file like or str, optional
            File to save model to.
        """
        if f is None:
            f = "model_" + datetime.datetime.now().isoformat() + ".pkl"
            if self.dataset is not None:
                f = "model_" + self.dataset.name + ".pkl"

        with possibly_open_file(f) as g:
            pickle.dump(self.model, g)

    def load_model(self, *args, f=None):
        """
        Load a serialised model.
        If `f` is not specified then an attempt will be made to find a model
        corresponding to the current dataset name,
        `'model_' + self.dataset.name + '.pkl'`. If there is no current
        dataset then the most recent model will be loaded.
        This method is only intended to be used to deserialise models created
        by this interactive Jupyter widget modeller, and will not successfully
        load complicated ReflectModel created outside of the interactive
        modeller.

        Parameters
        ----------
        f: file like or str, optional
            pickle file to load model from.
        """
        if f is None and self.dataset is not None:
            # try and load the model corresponding to the current dataset
            f = "model_" + self.dataset.name + ".pkl"
        elif f is None:
            # load the most recent model file
            files = list(filter(os.path.isfile, glob.glob("model_*.pkl")))
            files.sort(key=lambda x: os.path.getmtime(x))
            files.reverse()
            if len(files):
                f = files[0]

        if f is None:
            self._print("No model file is specified/available.")
            return

        try:
            with possibly_open_file(f, "rb") as g:
                reflect_model = pickle.load(g)
            self.set_model(reflect_model)
        except (RuntimeError, FileNotFoundError) as exc:
            # RuntimeError if the file isn't a ReflectModel
            # FileNotFoundError if the specified file name wasn't found
            self._print(repr(exc), repr(f))

    def set_model(self, model):
        """
        Change the `refnx.reflect.ReflectModel` associated with the `Motofit`
        instance.

        Parameters
        ----------
        model: refnx.reflect.ReflectModel

        """
        if not isinstance(model, ReflectModel):
            raise RuntimeError("`model` was not an instance of ReflectModel")

        if self.model_view is not None:
            self.model_view.unobserve_all()

        # figure out if the reflect_model is a different instance. If it is
        # then the objective has to be updated.
        if model is not self.model:
            self.model = model
            self._update_analysis_objects()

        self.model = model

        self.model_view = ReflectModelView(self.model)
        self.model_view.observe(self.update_model, names=["view_changed"])
        self.model_view.observe(self.redraw, names=["view_redraw"])

        # observe when the number of varying parameters changed. This
        # invalidates a curvefitter, and a new one has to be produced.
        self.model_view.observe(
            self._on_num_varying_changed, names=["num_varying"]
        )

        self.model_view.do_fit_button.on_click(self.do_fit)
        self.model_view.to_code_button.on_click(self._to_code)
        self.model_view.save_model_button.on_click(self.save_model)
        self.model_view.load_model_button.on_click(self.load_model)

        self.redraw(None)

    def update_model(self, change):
        """
        Updates the plots when the parameters change

        Parameters
        ----------
        change

        """
        if not self.fig:
            return

        q = np.linspace(self.qmin, self.qmax, self.qpnt)
        theoretical = self.model.model(q)
        yt, _ = self.transform(q, theoretical)

        sld_profile = self.model.structure.sld_profile()
        z, sld = sld_profile
        if self.theoretical_plot is not None:
            self.theoretical_plot.set_data(q, yt)

            self.theoretical_plot_sld.set_data(z, sld)
            self.ax_sld.relim()
            self.ax_sld.autoscale_view()

        if self.dataset is not None:
            # if there's a dataset loaded then residuals_plot
            # should exist
            residuals = self.objective.residuals()
            self.chisqr.value = np.sum(residuals**2)

            self.residuals_plot.set_data(self.dataset.x, residuals)
            self.ax_residual.relim()
            self.ax_residual.autoscale_view()

        self.fig.canvas.draw()

    def _on_num_varying_changed(self, change):
        # observe when the number of varying parameters changed. This
        # invalidates a curvefitter, and a new one has to be produced.
        if change["new"] != change["old"]:
            self._curvefitter = None

    def _update_analysis_objects(self):
        use_weights = self.use_weights.value == "Yes"
        self.objective = Objective(
            self.model,
            self.dataset,
            transform=self.transform,
            use_weights=use_weights,
        )
        self._curvefitter = None

    def __call__(self, data=None, model=None):
        """
        Display the `Motofit` GUI in a Jupyter notebook cell.

        Parameters
        ----------
        data: refnx.dataset.Data1D
            The dataset to associate with the `Motofit` instance.

        model: refnx.reflect.ReflectModel or str or file-like
            A model to associate with the data.
            If `model` is a `str` or `file`-like then the `load_model` method
            will be used to try and load the model from file. This assumes that
            the file is a pickle of a `ReflectModel`
        """
        # the theoretical model
        # display the main graph
        import matplotlib.pyplot as plt

        self.fig = plt.figure(figsize=(9, 4))

        # grid specs depending on whether the residuals are displayed
        if self.display_residuals.value:
            d_gs = self._gridspec1[0, 0]
            sld_gs = self._gridspec1[:, 1]
        else:
            d_gs = self._gridspec2[0, 0]
            sld_gs = self._gridspec2[0, 1]

        self.ax_data = self.fig.add_subplot(d_gs)
        self.ax_data.set_xlabel(r"$Q/\AA^{-1}$")
        self.ax_data.set_ylabel("Reflectivity")

        self.ax_data.grid(True, color="b", linestyle="--", linewidth=0.1)

        self.ax_sld = self.fig.add_subplot(sld_gs)
        self.ax_sld.set_ylabel(r"$\rho/10^{-6}\AA^{-2}$")
        self.ax_sld.set_xlabel(r"$z/\AA$")

        self.ax_residual = self.fig.add_subplot(
            self._gridspec1[1, 0], sharex=self.ax_data
        )
        self.ax_residual.set_xlabel(r"$Q/\AA^{-1}$")
        self.ax_residual.grid(True, color="b", linestyle="--", linewidth=0.1)
        self.ax_residual.set_visible(self.display_residuals.value)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.fig.tight_layout()

        q = np.linspace(self.qmin, self.qmax, self.qpnt)
        theoretical = self.model.model(q)
        yt, _ = self.transform(q, theoretical)

        self.theoretical_plot = self.ax_data.plot(q, yt, zorder=2)[0]
        self.ax_data.set_yscale("log")

        z, sld = self.model.structure.sld_profile()
        self.theoretical_plot_sld = self.ax_sld.plot(z, sld)[0]

        # the figure has been reset, so remove ref to the data_plot,
        # residual_plot
        self.data_plot = None
        self.residuals_plot = None

        self.dataset = None
        if data is not None:
            self.load_data(data)

        if isinstance(model, ReflectModel):
            self.set_model(model)
            return self.display_box
        elif model is not None:
            self.load_model(model)
            return self.display_box

        self.redraw(None)
        return self.display_box

    def load_data(self, data):
        """
        Load a dataset into the `Motofit` instance.

        Parameters
        ----------
        data: refnx.dataset.Data1D, or str, or file-like
        """
        if isinstance(data, ReflectDataset):
            self.dataset = data
        else:
            self.dataset = ReflectDataset(data)

        self.dataset_name.value = self.dataset.name

        # loading a dataset changes the objective and curvefitter
        self._update_analysis_objects()

        self.qmin = np.min(self.dataset.x)
        self.qmax = np.max(self.dataset.x)
        if self.fig is not None:
            yt, et = self.transform(self.dataset.x, self.dataset.y)

            if self.data_plot is None:
                (self.data_plot,) = self.ax_data.plot(
                    self.dataset.x,
                    yt,
                    label=self.dataset.name,
                    ms=2,
                    marker="o",
                    ls="",
                    zorder=1,
                )
                self.data_plot.set_label(self.dataset.name)
                self.ax_data.legend()

                # no need to calculate residuals here, that'll be updated in
                # the redraw method
                (self.residuals_plot,) = self.ax_residual.plot(self.dataset.x)
            else:
                self.data_plot.set_xdata(self.dataset.x)
                self.data_plot.set_ydata(yt)

            # calculate theoretical model over same range as data
            # use redraw over update_model because it ensures chi2 widget gets
            # displayed
            self.redraw(None)
            self.ax_data.relim()
            self.ax_data.autoscale_view()
            self.ax_residual.relim()
            self.ax_residual.autoscale_view()
            self.fig.canvas.draw()

    def redraw(self, change):
        """
        Redraw the Jupyter GUI associated with the `Motofit` instance.
        """
        self._update_display_box(self.display_box)
        self.update_model(None)

    @property
    def curvefitter(self):
        """
        class:`CurveFitter` : Object for fitting the data based on the
        objective.
        """
        if self.objective is not None and self._curvefitter is None:
            self._curvefitter = CurveFitter(self.objective)

        return self._curvefitter

    def _print(self, string):
        """
        Print to the output widget
        """
        from IPython.display import clear_output

        with self.output:
            clear_output()
            print(string)

    def do_fit(self, *args):
        """
        Ask the Motofit object to perform a fit (differential evolution).

        Parameters
        ----------
        change

        Notes
        -----
        After performing the fit the Jupyter display is updated.

        """
        if self.dataset is None:
            return

        if not self.model.parameters.varying_parameters():
            self._print("No parameters are being varied")
            return

        try:
            logp = self.objective.logp()
            if not np.isfinite(logp):
                self._print(
                    "One of your parameter values lies outside its"
                    " bounds. Please adjust the value, or the bounds."
                )
                return
        except ZeroDivisionError:
            self._print(
                "One parameter has equal lower and upper bounds."
                " Either alter the bounds, or don't let that"
                " parameter vary."
            )
            return

        def callback(xk, convergence):
            self.chisqr.value = self.objective.chisqr(xk)

        self.curvefitter.fit("differential_evolution", callback=callback)

        # need to update the widgets as the model will be updated.
        # this also redraws GUI.
        # self.model_view.refresh()
        self.set_model(self.model)

        self._print(str(self.objective))

    def _to_code(self, change=None):
        self._print(self.code)

    @property
    def code(self):
        """
        str : A Python code fragment capable of fitting the data.
        Executable Python code fragment for the GUI model.
        """
        if self.objective is None:
            self._update_analysis_objects()

        return to_code(self.objective)

    def _on_tab_changed(self, change):
        pass

    def _on_plot_type_changed(self, change):
        """
        User would like to plot and fit as logR/linR/RQ4/RQ2, etc
        """
        self.transform = Transform(change["new"])
        if self.objective is not None:
            self.objective.transform = self.transform

        if self.dataset is not None:
            yt, _ = self.transform(self.dataset.x, self.dataset.y)

            self.data_plot.set_xdata(self.dataset.x)
            self.data_plot.set_ydata(yt)

        self.update_model(None)

        # probably have to change LHS axis of the data plot when
        # going between different plot types.
        if change["new"] == "logY":
            self.ax_data.set_yscale("linear")
        else:
            self.ax_data.set_yscale("log")

        self.ax_data.relim()
        self.ax_data.autoscale_view()
        self.fig.canvas.draw()

    def _on_use_weights_changed(self, change):
        self._update_analysis_objects()
        self.update_model(None)

    def _on_display_residuals_changed(self, change):
        import matplotlib.pyplot as plt

        if change["new"]:
            self.ax_residual.set_visible(True)
            self.ax_data.set_position(
                self._gridspec1[0, 0].get_position(self.fig)
            )
            self.ax_sld.set_position(
                self._gridspec1[:, 1].get_position(self.fig)
            )
            plt.setp(self.ax_data.get_xticklabels(), visible=False)
        else:
            self.ax_residual.set_visible(False)
            self.ax_data.set_position(
                self._gridspec2[:, 0].get_position(self.fig)
            )
            self.ax_sld.set_position(
                self._gridspec2[:, 1].get_position(self.fig)
            )
            plt.setp(self.ax_data.get_xticklabels(), visible=True)

    @property
    def _options_box(self):
        return widgets.VBox(
            [self.plot_type, self.use_weights, self.display_residuals]
        )

    def _update_display_box(self, box):
        """
        Redraw the Jupyter GUI associated with the `Motofit` instance
        """
        vbox_widgets = []

        if self.dataset is not None:
            vbox_widgets.append(widgets.HBox([self.dataset_name, self.chisqr]))

        self.tab.children = [
            self.model_view.model_box,
            self.model_view.limits_box,
            self._options_box,
        ]

        vbox_widgets.append(self.tab)
        vbox_widgets.append(self.output)
        box.children = tuple(vbox_widgets)


def rename_params(structure):
    for i in range(1, len(structure) - 1):
        structure[i].thick.name = "%d - thick" % i
        structure[i].sld.real.name = "%d - sld" % i
        structure[i].sld.imag.name = "%d - isld" % i
        structure[i].rough.name = "%d - rough" % i

    structure[0].sld.real.name = "sld - fronting"
    structure[0].sld.imag.name = "isld - fronting"
    structure[-1].sld.real.name = "sld - backing"
    structure[-1].sld.imag.name = "isld - backing"
    structure[-1].rough.name = "rough - backing"


def to_code(objective):
    """
    Create executable Python code fragment that corresponds to the model in the
    GUI.

    Parameters
    ----------
    objective: refnx.analysis.Objective

    Returns
    -------
    code: str
        Python code that can construct a reflectometry fitting system

    """
    header = """import numpy as np
import refnx
from refnx.analysis import Objective, CurveFitter, Transform
from refnx.reflect import SLD, Slab, ReflectModel, Structure
from refnx.dataset import ReflectDataset

print(refnx.version.version)
    """

    code = [header]

    # the dataset
    code.append('data = ReflectDataset("{0}")'.format(objective.data.filename))

    # make some SLD's and slabs
    slds = ["\n# set up the SLD objects for each layer"]
    slabs = ["\n# set up the Slab objects for each layer"]
    limits = []

    structure = "structure = "
    for i, slab in enumerate(objective.model.structure):
        sld = slab.sld
        lims = [
            (sld.real, "sld{0}.real"),
            (sld.imag, "sld{0}.imag"),
            (slab.thick, "slab{0}.thick"),
            (slab.rough, "slab{0}.rough"),
        ]

        slds.append(
            "sld{0} = SLD({1} + {2}j, name='{3}')".format(
                i, sld.real.value, sld.imag.value, slab.name
            )
        )

        slabs.append(
            "slab{0} = Slab({1}, sld{0}, {2}, name='{3}')".format(
                i, slab.thick.value, slab.rough.value, slab.name
            )
        )

        for p, temp in lims:
            if p.vary:
                limits.append(
                    (temp + ".setp(vary=True, bounds=({1}, {2}))").format(
                        i, p.bounds.lb, p.bounds.ub
                    )
                )

        if not i:
            structure += "slab{0}".format(i)
        else:
            structure += " | slab{0}".format(i)

    code.extend(slds)
    code.extend(slabs)

    code.append("\n# set up the limits for SLD's and Slabs")
    code.extend(limits)

    code.append("\n# set up the Structure object from the Slabs")
    code.append(structure)

    model = objective.model
    code.append("\n# make the reflectometry model")
    code.append(
        "model = ReflectModel(structure, scale={0}, bkg={1}, dq={2})".format(
            model.scale.value, model.bkg.value, model.dq.value
        )
    )

    lims = [
        (model.scale, "model.scale"),
        (model.bkg, "model.bkg"),
        (model.dq, "model.dq"),
    ]

    for p, temp in lims:
        if p.vary:
            code.append(
                (temp + ".setp(vary=True, bounds=({0}, {1}))").format(
                    p.bounds.lb, p.bounds.ub
                )
            )

    code.append("\n# make the objective")
    code.append("transform = {}".format(repr(objective.transform)))
    code.append(
        "objective = Objective(model, data, transform=transform,"
        " use_weights={})".format(objective.weighted)
    )

    code.append("\n# make the curvefitter")
    code.append("fitter = CurveFitter(objective)")
    code.append("fitter.fit('differential_evolution')")
    code.append("print(objective)")

    return "\n".join(code)



""""
A basic representation of a 1D dataset
"""
import os.path
import re

import numpy as np
from scipy._lib._util import check_random_state
from refnx.util.nsplice import get_scaling_in_overlap
from refnx._lib import possibly_open_file


class Data1D:
    r"""
    A basic representation of a 1D dataset.

    Parameters
    ----------
    data : str, file-like or tuple of np.ndarray, optional
        `data` can be a string or file-like object referring to a File to load
        the dataset from. The file should be plain text and have 2 to 4
        columns separated by space, comma or tab. The columns represent
        `x, y [y_err [, x_err]]`.

        Alternatively it is a tuple containing the data from which the dataset
        will be constructed. The tuple should have between 2 and 4 members.

            - data[0] - x
            - data[1] - y
            - data[2] - uncertainties on y, y_err
            - data[3] - uncertainties on x, x_err

        `data` must be at least two long, `x` and `y`.
        If the tuple is at least 3 long then the third member is `y_err`.
        If the tuple is 4 long then the fourth member is `x_err`.
        All arrays must have the same shape.

    mask : array-like
        Specifies which data points are (un)masked. Must be broadcastable
        to the y-data. `Data1D.mask = None` clears the mask. If a mask value
        equates to `True`, then the point is included, if a mask value equates
        to `False` it is excluded.

    Attributes
    ----------
    filename : str or None
        The file the data was read from
    weighted : bool
        Whether the y data has uncertainties
    metadata : dict
        Information that should be retained with the dataset.

    """

    def __init__(self, data=None, mask=None, **kwds):
        self.filename = None
        self.name = None

        self.metadata = kwds
        self._x = np.zeros(0)
        self._y = np.zeros(0)
        self._y_err = None
        self._x_err = None
        self.weighted = False

        # if it's a file then open and load the file.
        if hasattr(data, "read") or type(data) is str:
            self.load(data)
        elif isinstance(data, Data1D):
            # copy a dataset (but not it's file info)
            self.name = data.name
            self.filename = data.filename
            self.metadata = data.metadata
            self._x = data._x
            self._y = data._y
            self._y_err = data._y_err
            self._x_err = data._x_err
            self.weighted = data.weighted
            self._mask = data._mask
        elif data is not None:
            self._x = np.array(data[0], dtype=float)
            self._y = np.array(data[1], dtype=float)
            if len(data) > 2:
                self._y_err = np.array(data[2], dtype=float)
                self.weighted = True

            if len(data) > 3:
                self._x_err = np.array(data[3], dtype=float)

        self._mask = None
        if mask is not None:
            self._mask = np.broadcast_to(mask, self._y.shape)

    def __len__(self):
        """
        the number of unmasked points in the dataset.

        """
        return self.y.size

    def __str__(self):
        return "<{0}>, {1} points".format(self.name, len(self))

    def __repr__(self):
        msk = self._mask
        if np.all(self._mask):
            msk = None

        d = {"filename": self.filename, "msk": msk, "data": self.data}
        if self.filename is not None:
            return "Data1D(data={filename!r}," " mask={msk!r})".format(**d)
        else:
            return "Data1D(data={data!r}," " mask={msk!r})".format(**d)

    @property
    def x(self):
        """
        np.ndarray : x data (possibly masked)
        """
        if self._mask is not None and self._x.size:
            return self._x[self.mask]
        return self._x

    @property
    def y(self):
        """
        np.ndarray : y data (possibly masked)
        """
        if self._mask is not None and self._y.size:
            return self._y[self.mask]
        return self._y

    @property
    def x_err(self):
        """
        np.ndarray : x uncertainty (possibly masked)
        """
        if (
            self._x_err is not None
            and self._mask is not None
            and self._x_err.size
        ):
            return self._x_err[self.mask]
        return self._x_err

    @x_err.setter
    def x_err(self, x_err):
        """
        np.ndarray : y uncertainty (possibly masked)
        """
        self._x_err = x_err

    @property
    def y_err(self):
        """
        uncertainties on the y data (possibly masked)
        """
        if (
            self._y_err is not None
            and self._mask is not None
            and self._y_err.size
        ):
            return self._y_err[self.mask]
        return self._y_err

    @property
    def mask(self):
        """
        mask
        """
        if self._mask is None:
            self._mask = np.full_like(self._y, True, dtype=bool)

        return self._mask

    @mask.setter
    def mask(self, mask):
        """
        mask
        """
        if mask is None:
            mask = True

        self._mask = np.broadcast_to(mask, self._y.shape).astype(bool)

    @property
    def data(self):
        """
        4-tuple containing the (`x`, `y`, `y_err`, `x_err`) data

        """
        return self.x, self.y, self.y_err, self.x_err

    @property
    def unmasked_data(self):
        """
        4-tuple containing unmasked (x, y, y_err, x_err) data

        """
        return self._x, self._y, self._y_err, self._x_err

    @property
    def finite_data(self):
        """
        4-tuple containing the (`x`, `y`, `y_err`, `x_err`) datapoints that are
        finite.

        """
        finite_loc = np.where(np.isfinite(self.y))
        return (
            self.x[finite_loc],
            self.y[finite_loc],
            self.y_err[finite_loc],
            self.x_err[finite_loc],
        )

    @data.setter
    def data(self, data_tuple):
        """
        Set the data for this object from supplied data.

        Parameters
        ----------
        data_tuple : tuple
            2 to 4 member tuple containing the (x, y, y_err, x_err) data to
            specify the dataset. `y_err` and `x_err` are optional.

        Notes
        -----
        Clears the mask for the dataset, it will need to be reapplied.

        """
        self._x = np.array(data_tuple[0], dtype=float)
        self._y = np.array(data_tuple[1], dtype=float)
        self.weighted = False
        self._y_err = None
        self._x_err = None

        if len(data_tuple) > 2 and data_tuple[2] is not None:
            self._y_err = np.array(data_tuple[2], dtype=float)
            self.weighted = True

        if len(data_tuple) > 3 and data_tuple[3] is not None:
            self._x_err = np.array(data_tuple[3], dtype=float)

        self._mask = None
        self.sort()

    def scale(self, scalefactor=1.0):
        """
        Scales the y and y_err data by dividing by `scalefactor`.

        Parameters
        ----------
        scalefactor : float
            The scalefactor to divide by.

        """
        self._y /= scalefactor
        self._y_err /= scalefactor

    def add_data(self, data_tuple, requires_splice=False, trim_trailing=True):
        """
        Adds more data to the dataset.

        Parameters
        ----------
        data_tuple : tuple
            2 to 4 member tuple containing the (x, y, y_err, x_err) data to add
            to the dataset. `y_err` and `x_err` are optional.
        requires_splice : bool, optional
            When the new data is added to the dataset do you want to scale it
            vertically so that it overlaps with the existing data? `y` and
            `y_err` in `data_tuple` are both multiplied by the scaling factor.
        trim_trailing : bool, optional
            When the new data is concatenated do you want to remove points from
            the existing data that are in the overlap region? This might be
            done because the datapoints in the `data_tuple` you are adding have
            have lower `y_err` than the preceding data.

        Notes
        -----
        Raises `ValueError` if there are no points in the overlap region and
        `requires_splice` was True. The added data is not masked.

        """
        x, y, y_err, x_err = self.data

        # dataset has no points, can just initialise with the tuple
        if not len(self):
            self.data = data_tuple
            return

        ax, ay = data_tuple[0:2]

        # if ((len(data_tuple) > 2 and self.y_err is None) or
        #         (len(data_tuple) == 2 and self.y_err is not None)):
        #     raise ValueError("Both the existing Data1D and the data you're"
        #                      " trying to add need to have y_err")
        #
        # if ((len(data_tuple) > 3 and self.x_err is None) or
        #         (len(data_tuple) == 3 and self.x_err is not None)):
        #     raise ValueError("Both the existing Data1D and the data you're"
        #                      " trying to add need to have x_err")

        ay_err = None
        ax_err = None

        if len(data_tuple) > 2:
            ay_err = np.array(data_tuple[2], dtype=float)

        if len(data_tuple) > 3:
            ax_err = np.array(data_tuple[3], dtype=float)

        mask2 = np.full_like(data_tuple[0], True, bool)

        # which values in the first dataset overlap with the second
        overlap_points = np.zeros_like(x, "bool")

        # go through and stitch them together.
        scale = 1.0
        dscale = 0.0
        if requires_splice and len(self) > 1:
            scale, dscale, overlap_points = get_scaling_in_overlap(
                x, y, y_err, ax, ay, ay_err
            )

            if (
                (not np.isfinite(scale))
                or (not np.isfinite(dscale))
                or (not np.size(overlap_points, 0))
            ):
                raise ValueError("No points in overlap region")

        if not trim_trailing:
            overlap_points[:] = False

        qq = np.r_[x[~overlap_points], ax]
        rr = np.r_[y[~overlap_points], ay * scale]
        overall_mask = np.r_[self.mask[~overlap_points], mask2]

        try:
            dr = np.r_[y_err[~overlap_points], ay_err * scale]
        except (TypeError, ValueError):
            if (ay_err is not None) or (y_err is not None):
                raise ValueError(
                    "Both the existing Data1D and the data you're"
                    " trying to add need to have y_err"
                )
            dr = None

        try:
            dq = np.r_[x_err[~overlap_points], ax_err]
        except (TypeError, ValueError):
            if (ax_err is not None) or (x_err is not None):
                raise ValueError(
                    "Both the existing Data1D and the data you're"
                    " trying to add need to have x_err"
                )
            dq = None

        self.data = (qq, rr, dr, dq)
        self.mask = overall_mask

        self.sort()

    def sort(self):
        """
        Sorts the data in ascending order
        """
        sorted = np.argsort(self.x)
        self._x = self.x[sorted]
        self._y = self.y[sorted]

        if self._mask is not None:
            self._mask = self._mask[sorted]
        if self.y_err is not None:
            self._y_err = self.y_err[sorted]
        if self.x_err is not None:
            self._x_err = self.x_err[sorted]

    def save(self, f, header=None):
        """
        Saves the data to file. Saves the data as 4 column ASCII.

        Parameters
        ----------
        f : file-handle or string
            File to save the dataset to.

        """
        if header is not None:
            _header = header
        elif len(self.metadata):
            _header = str(self.metadata)
        else:
            _header = ""

        np.savetxt(
            f,
            np.column_stack((self._x, self._y, self._y_err, self._x_err)),
            header=_header,
        )

    def load(self, f):
        """
        Loads a dataset from file, and overwrites existing data.
        Must be 2 to 4 column ASCII.

        Parameters
        ----------
        f : file-handle or string
            File to load the dataset from.

        """
        # it would be nicer to simply use np.loadtxt, but this is an
        # attempt to auto ignore header lines.
        with possibly_open_file(f, "r") as g:
            lines = list(reversed(g.readlines()))
            x = list()
            y = list()
            y_err = list()
            x_err = list()

            # a marker for how many columns in the data there will be
            numcols = 0
            for i, line in enumerate(lines):
                try:
                    # parse a line for numerical tokens separated by whitespace
                    # or comma
                    nums = [
                        float(tok)
                        for tok in re.split(r"\s|,", line)
                        if len(tok)
                    ]
                    if len(nums) in [0, 1]:
                        # might be trailing newlines at the end of the file,
                        # just ignore those
                        continue
                    if not numcols:
                        # figure out how many columns one has
                        numcols = len(nums)
                    elif len(nums) != numcols:
                        # if the number of columns changes there's an issue
                        break
                    x.append(nums[0])
                    y.append(nums[1])
                    if len(nums) > 2:
                        y_err.append(nums[2])
                    if len(nums) > 3:
                        x_err.append(nums[3])
                except ValueError:
                    # you should drop into this if you can't parse tokens into
                    # a series of floats. But the text may be meta-data, so
                    # try to carry on.
                    continue

        x.reverse()
        y.reverse()
        y_err.reverse()
        x_err.reverse()

        if len(x) == 0:
            raise RuntimeError(
                "Datafile didn't appear to contain any data (or"
                " was the wrong format)"
            )

        if numcols < 3:
            y_err = None
        if numcols < 4:
            x_err = None

        self.data = (x, y, y_err, x_err)

        if hasattr(f, "read"):
            fname = f.name
        else:
            fname = f

        self.filename = fname
        self.name = os.path.splitext(os.path.basename(fname))[0]

    def refresh(self):
        """
        Refreshes a previously loaded dataset.

        """
        if self.filename is not None:
            with open(self.filename) as f:
                self.load(f)

    def plot(self, fig=None):
        """
        Plot the dataset.

        Requires matplotlib be installed.

        Parameters
        ----------
        fig: Figure instance, optional
            If `fig` is not supplied then a new figure is created. Otherwise
            the graph is created on the current axes on the supplied figure.

        Returns
        -------
        fig, ax : :class:`matplotlib.figure.Figure`, :class:`matplotlib.Axes`
            `matplotlib` figure and axes objects.

        """
        import matplotlib.pyplot as plt

        if fig is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        else:
            ax = fig.gca()

        if self.y_err is not None:
            ax.errorbar(self.x, self.y, self.y_err, label=self.name)
        else:
            ax.scatter(self.x, self.y, label=self.name)

        return fig, ax

    def __add__(self, other):
        """
        Adds two datasets together. Splices the data and trims data in the
        overlap region.

        """
        ret = Data1D(self.data)
        ret.add_data(other.data, requires_splice=True, trim_trailing=True)
        return ret

    def __radd__(self, other):
        """
        radd of two datasets. Splices the data and trims data in the
        overlap region.

        """
        self.add_data(other.data, requires_splice=True, trim_trailing=True)
        return self

    def __iadd__(self, other):
        """
        iadd of two datasets. Splices the data and trims data in the
        overlap region.

        """
        self.add_data(other.data, requires_splice=True, trim_trailing=True)
        return self

    def synthesise(self, random_state=None):
        """
        Synthesise a new dataset by adding Gaussian noise onto each of the
        datapoints of the existing data.

        Returns
        -------
        dataset : :class:`Data1D`
            A new synthesised dataset
        random_state : {int, :class:`numpy.random.RandomState`, :class:`numpy.random.Generator`}
            If `random_state` is not specified the
            :class:`numpy.random.RandomState` singleton is used.
            If `random_state` is an int, a new ``RandomState`` instance is
            used, seeded with random_state.
            If `random_state` is already a ``RandomState`` or a ``Generator``
            instance, then that object is used.
            Specify `random_state` for repeatable synthesising.
        """
        if self._y_err is None:
            raise RuntimeError(
                "Can't synthesise new dataset without y_err" "uncertainties"
            )

        rng = check_random_state(random_state)
        gnoise = rng.standard_normal(size=self._y_err.shape)

        new_y = self._y + gnoise * self._y_err
        data = list(self.data)
        data.pop(1)
        data.insert(1, new_y)

        dataset = Data1D()
        dataset.data = data

        return dataset



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

__all__ = ['Chain']

import attr
import numpy as np

from . import util
from . import ensemble


@attr.s(slots=True)
class Chain(object):
    ensemble = attr.ib(type=ensemble.Ensemble)
    thin_by = attr.ib(type=int, default=None)
    x = attr.ib(type=np.ndarray, init=False, default=None)
    logP = attr.ib(type=np.ndarray, init=False, default=None)
    logl = attr.ib(type=np.ndarray, init=False, default=None)
    betas = attr.ib(type=np.ndarray, init=False, default=None)

    swaps_proposed = attr.ib(type=np.ndarray, init=False)
    swaps_accepted = attr.ib(type=np.ndarray, init=False)
    jumps_proposed = attr.ib(type=np.ndarray, init=False)
    jumps_accepted = attr.ib(type=np.ndarray, init=False)

    def __attrs_post_init__(self):
        if self.thin_by is None:
            self.thin_by = 1
        self.x = np.empty((0, self.ntemps, self.nwalkers, self.ndim), float)
        self.logP = np.empty((0, self.ntemps, self.nwalkers), float)
        self.logl = np.empty((0, self.ntemps, self.nwalkers), float)
        self.betas = np.empty((0, self.ntemps), float)

        self.jumps_proposed = np.zeros((self.ntemps, self.nwalkers))
        self.jumps_accepted = np.zeros((self.ntemps, self.nwalkers))
        self.swaps_proposed = np.zeros(self.ntemps - 1)
        self.swaps_accepted = np.zeros(self.ntemps - 1)

    @property
    def length(self):
        return self.x.shape[0]

    @property
    def time(self):
        return self.ensemble.time

    @property
    def ntemps(self):
        return self.ensemble.ntemps

    @property
    def nwalkers(self):
        return self.ensemble.nwalkers

    @property
    def ndim(self):
        return self.ensemble.ndim

    @property
    def jump_acceptance_ratio(self):
        return self.jumps_accepted / self.jumps_proposed

    @property
    def swap_acceptance_ratio(self):
        return self.swaps_accepted / self.swaps_proposed

    @staticmethod
    def _resize(array, count):
        shape = (count,) + array.shape[1:]
        return np.concatenate((array, np.empty(shape)), axis=0)

    def run(self, count):
        jp0 = self.jumps_proposed.copy()
        ja0 = self.jumps_accepted.copy()
        sp0 = self.swaps_proposed.copy()
        sa0 = self.swaps_accepted.copy()
        for _ in self.iterate(count):
            pass
        jp = self.jumps_proposed - jp0
        ja = self.jumps_accepted - ja0
        sp = self.swaps_proposed - sp0
        sa = self.swaps_accepted - sa0
        return ja / jp, sa / sp

    def iterate(self, count):
        start = self.length
        self.x = self._resize(self.x, count)
        self.logP = self._resize(self.logP, count)
        self.logl = self._resize(self.logl, count)
        self.betas = self._resize(self.betas, count)
        for i in range(start, start + count):
            for _ in range(self.thin_by):
                self.ensemble.step()
                self.swaps_proposed += self.ensemble.swaps_proposed
                self.swaps_accepted += self.ensemble.swaps_accepted
                self.jumps_proposed += self.ensemble.jumps_proposed
                self.jumps_accepted += self.ensemble.jumps_accepted

            self.x[i] = self.ensemble.x
            self.logP[i] = self.ensemble.logP
            self.logl[i] = self.ensemble.logl
            self.betas[i] = self.ensemble.betas
            yield self.ensemble

    def get_acts(self, window=50):
        acts = np.zeros((self.ntemps, self.ndim))

        for i in range(self.ntemps):
            x = np.mean(self.x[:, i, :, :], axis=1)
            acts[i, :] = util.get_integrated_act(x, window=window)
        return acts

    def log_evidence_estimate(self, fburnin=0.1):
        """
        Thermodynamic integration estimate of the evidence for the sampler.

        :param fburnin: (optional)
            The fraction of the chain to discard as burnin samples; only the
            final ``1-fburnin`` fraction of the samples will be used to
            compute the evidence; the default is ``fburnin = 0.1``.
        :return ``(logZ, dlogZ)``: Returns an estimate of the
            log-evidence and the error associated with the finite
            number of temperatures at which the posterior has been
            sampled.
        For details, see ``thermodynamic_integration_log_evidence``.

        """

        istart = int(self.logl.shape[0] * fburnin + 0.5)
        mean_logls = np.mean(np.mean(self.logl, axis=2)[istart:, :], axis=0)
        return util.thermodynamic_integration_log_evidence(self.betas[-1], mean_logls)



import sys
import logging
import time
import os.path


def gui():
    from PyQt5 import QtWidgets
    from refnx.reduce._app.view import SlimWindow

    time_str = time.strftime("%Y%m%d-%H%M%S")
    log_filename = "slim_" + time_str + ".log"
    log_filename = os.path.join(os.path.expanduser("~"), log_filename)

    logging.basicConfig(
        filename=log_filename,
        level=logging.INFO,
        format="%(asctime)s %(message)s",
    )
    logging.info("Starting SLIM reduction")

    app = QtWidgets.QApplication(sys.argv)
    myapp = SlimWindow()

    myapp.show()
    v = app.exec_()
    return v


def main():
    sys.exit(gui())


__all__ = [gui, main]



"""numerical differentiation function, gradient, Jacobian, and Hessian
Author : josef-pkt
License : BSD
Notes
-----
These are simple forward differentiation, so that we have them available
without dependencies.
* Jacobian should be faster than numdifftools because it doesn't use loop over
  observations.
* numerical precision will vary and depend on the choice of stepsizes

VENDORED Aug2017
"""
import numpy as np


# NOTE: we only do double precision internally so far
EPS = np.finfo(np.float64).eps


def _get_epsilon(x, s, epsilon, n):
    if epsilon is None:
        h = EPS ** (1.0 / s) * np.maximum(np.abs(x), 0.1)
    else:
        if np.isscalar(epsilon):
            h = np.empty(n)
            h.fill(epsilon)
        else:  # pragma : no cover
            h = np.asarray(epsilon)
            if h.shape != x.shape:
                raise ValueError(
                    "If h is not a scalar it must have the same" " shape as x."
                )
    return h


def approx_hess2(x, f, epsilon=None, args=(), kwargs={}, return_grad=False):
    """
    Calculate Hessian with finite difference derivative approximation
    Parameters
    ----------
    x : array_like
       value at which function derivative is evaluated
    f : function
       function of one array f(x, `*args`, `**kwargs`)
    epsilon : float or array-like, optional
       Stepsize used, if None, then stepsize is automatically chosen
       according to EPS**(1/3)*x.
    args : tuple
        Arguments for function `f`.
    kwargs : dict
        Keyword arguments for function `f`.

    Returns
    -------
    hess : ndarray
       array of partial second derivatives, Hessian

    Notes
    -----
    Equation (%(equation_number)s) in Ridout. Computes the Hessian as::
      %(equation)s
    where e[j] is a vector with element j == 1 and the rest are zero and
    d[i] is epsilon[i].
    References
    ----------:
    Ridout, M.S. (2009) Statistical applications of the complex-step method
        of numerical differentiation. The American Statistician, 63, 66-74
    """
    n = len(x)
    # NOTE: ridout suggesting using eps**(1/4)*theta
    h = _get_epsilon(x, 3, epsilon, n)
    ee = np.diag(h)
    f0 = f(*((x,) + args), **kwargs)
    # Compute forward step
    g = np.zeros(n)
    gg = np.zeros(n)
    for i in range(n):
        g[i] = f(*((x + ee[i, :],) + args), **kwargs)
        gg[i] = f(*((x - ee[i, :],) + args), **kwargs)

    hess = np.outer(h, h)  # this is now epsilon**2
    # Compute "double" forward step
    for i in range(n):
        for j in range(i, n):
            hess[i, j] = (
                f(*((x + ee[i, :] + ee[j, :],) + args), **kwargs)
                - g[i]
                - g[j]
                + f0
                + f(*((x - ee[i, :] - ee[j, :],) + args), **kwargs)
                - gg[i]
                - gg[j]
                + f0
            ) / (2 * hess[i, j])
            hess[j, i] = hess[i, j]
    if return_grad:
        grad = (g - f0) / h
        return hess, grad
    else:
        return hess



from copy import deepcopy
import os.path
import pickle
from operator import itemgetter

import numpy as np
from PyQt5 import QtCore, QtGui, QtWidgets, uic

from refnx._lib import flatten, unique
from refnx.analysis import Parameter, Parameters, possibly_create_parameter
from refnx.dataset import ReflectDataset
from refnx.reflect._app.dataobject import DataObject
from refnx.reflect._app.datastore import DataStore
from refnx.reflect import (
    Slab,
    LipidLeaflet,
    SLD,
    ReflectModel,
    MixedReflectModel,
    Spline,
    Stack,
)


def component_class(component):
    """
    Give the node type for a given component
    """
    if isinstance(component, Slab):
        return SlabNode
    elif isinstance(component, LipidLeaflet):
        return LipidLeafletNode
    elif isinstance(component, Spline):
        return SplineNode
    elif isinstance(component, Stack):
        return StackNode


class Node:
    def __init__(self, data, model, parent=QtCore.QModelIndex()):
        self._data = data
        self._children = []
        # the parent node
        self._parent = parent

        # the qabstractitemmodel
        self._model = model

    def insertChild(self, i, item):
        item._model = self._model
        self._children.insert(i, item)

    def appendChild(self, item):
        item._model = self._model
        self._children.append(item)

    def popChild(self, row):
        return self._children.pop(row)

    def child(self, row):
        return self._children[row] if row < len(self._children) else None

    def childCount(self):
        return len(self._children)

    def columnCount(self):
        # name, value, sigma, lower, upper, expr
        # vary will be displayed in a checkbox with value
        return 6

    def data(self, column, role=QtCore.Qt.DisplayRole):
        return None

    def setData(self, col, val, role=QtCore.Qt.EditRole):
        # TreeModel emits the dataChanged Signal, not the node.
        return False

    def flags(self, column):
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        return flags

    def parent(self):
        return self._parent

    def row(self):
        if self._parent:
            # this index method is searching a *list* for an item, and isn't
            # meant to produce a QModelIndex
            v = self._parent._children.index(self)
            return v

        return 0

    def _row_parent(self):
        """
        Work out which row of the parent the current node lies in,
        returning the row and the parent.
        """
        return self.row(), self._parent

    def hierarchy(self):
        """
        The top down node hierarchy up to, and including, this node.
        """
        h = [self]
        rp = self._row_parent()
        while rp[1] is not None:
            h.append(rp[1])
            rp = rp[1]._row_parent()
        h.reverse()
        return h

    def row_indices(self):
        hierarchy = self.hierarchy()
        indices = []
        for node in hierarchy:
            indices.append(node.row())
        return indices

    def descendants(self):
        # yield all of the descendants of this node
        for el in self._children:
            yield el
            if el._children:
                yield from el.descendants()

    @property
    def index(self):
        """
        The QModelIndex for this node
        """
        # first work out the hierarchy of the nodes right back to the root
        # node, working out which row of the parent each node is in.
        # uses recursion
        a = []
        rp = self._row_parent()
        while rp[1] is not None:
            a.append(rp)
            rp = rp[1]._row_parent()

        # once you've got the hierarchy then you have to reverse it, and create
        # a model index right back down again.
        a.reverse()
        idx = QtCore.QModelIndex()
        for v in a:
            idx = self._model.index(v[0], 0, idx)
        return idx


class ParNode(Node):
    def __init__(self, data, model, parent=QtCore.QModelIndex()):
        super().__init__(data, model, parent)

    @property
    def parameter(self):
        return self._data

    def flags(self, column):
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        if column in [1, 3, 4]:
            flags |= QtCore.Qt.ItemIsEditable

        if column == 1:
            flags |= QtCore.Qt.ItemIsUserCheckable

        return flags

    def data(self, column, role=QtCore.Qt.DisplayRole):
        if role == QtCore.Qt.CheckStateRole:
            if column == 1:
                p = self.parameter
                if p.vary:
                    return QtCore.Qt.Checked
                else:
                    return QtCore.Qt.Unchecked
            else:
                return None
        elif role == QtCore.Qt.ToolTipRole:
            if column in [1, 2]:
                p = self.parameter
                return getattr(p, "units", None)
            elif column == 3:
                return "Lower limit for the parameter"
            elif column == 4:
                return "Upper limit for the parameter"

        if role == QtCore.Qt.DisplayRole:
            p = self.parameter
            d = [
                p.name,
                float(p.value),
                p.stderr,
                p.bounds.lb,
                p.bounds.ub,
                repr(p.constraint),
            ]
            return d[column]

    def setData(self, column, value, role=QtCore.Qt.EditRole):
        # we want to use a checkbox to say if a parameter is varying
        if role == QtCore.Qt.CheckStateRole and column == 1:
            try:
                p = self.parameter
                p.vary = value == QtCore.Qt.Checked
            except RuntimeError:
                # can't try and hold a parameter that has a constraint
                False

            return True

        # parse and fill out parameter values/limits
        if role == QtCore.Qt.EditRole:
            if type(value) is str:
                validator = QtGui.QDoubleValidator()
                voutput = validator.validate(value, 1)
            else:
                voutput = [QtGui.QValidator.Acceptable, float(value)]

            if voutput[0] == QtGui.QValidator.Acceptable:
                p = self.parameter
                if column == 1:
                    p.value = float(voutput[1])
                elif column == 3:
                    p.bounds.lb = float(voutput[1])
                elif column == 4:
                    p.bounds.ub = float(voutput[1])
            else:
                return False

        return True


class ParametersNode(Node):
    def __init__(self, data, model, parent=QtCore.QModelIndex()):
        super().__init__(data, model, parent)
        for p in data:
            if isinstance(p, Parameters):
                n = ParametersNode(p, model, self)
            if isinstance(p, Parameter):
                n = ParNode(p, model, self)
            self.appendChild(n)

    @property
    def parameters(self):
        return self._data

    def flags(self, column):
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled

        return flags

    def data(self, column, role=QtCore.Qt.DisplayRole):
        if role == QtCore.Qt.DisplayRole:
            if column == 0:
                name = self._data.name
                if not name:
                    name = "Parameters"
                return name

        return None


class PropertyNode(Node):
    # an object that displays/edits some attribute of its parent node
    # it is not a ParNode.
    def __init__(
        self, data, model, parent=QtCore.QModelIndex(), validators=()
    ):
        super().__init__(data, model, parent)
        # here self._data is the attribute name
        self.attribute_type = type(getattr(parent._data, data))
        self.validators = validators

    def flags(self, column):
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        if column == 1:
            flags |= QtCore.Qt.ItemIsEditable

            if self.attribute_type is bool:
                flags |= QtCore.Qt.ItemIsUserCheckable

        return flags

    def data(self, column, role=QtCore.Qt.DisplayRole):
        if (
            role == QtCore.Qt.CheckStateRole
            and column == 1
            and self.attribute_type is bool
        ):
            d = getattr(self._parent._data, self._data)

            if d:
                return QtCore.Qt.Checked
            else:
                return QtCore.Qt.Unchecked

        if role == QtCore.Qt.DisplayRole and column == 1:
            return getattr(self._parent._data, self._data)
        if role == QtCore.Qt.DisplayRole and column == 0:
            return self._data

    def columnCount(self):
        return 2

    def setData(self, column, value, role=QtCore.Qt.EditRole):
        if (
            role == QtCore.Qt.CheckStateRole
            and column == 1
            and self.attribute_type is bool
        ):

            if value == QtCore.Qt.Checked:
                setattr(self._parent._data, self._data, True)
            else:
                setattr(self._parent._data, self._data, False)
            return True

        # parse and fill out parameter values/limits
        if role == QtCore.Qt.EditRole and len(self.validators) and column == 1:
            for validator in self.validators:
                voutput = validator.validate(value, 1)
                if voutput[0] == QtGui.QValidator.Acceptable:
                    setattr(
                        self._parent._data,
                        self._data,
                        self.attribute_type(voutput[1]),
                    )
                    return True

        return False


class ComponentNode(Node):
    def __init__(self, data, model, parent=QtCore.QModelIndex(), flat=True):
        """
        Parameters
        ----------
        flat : bool
            If `flat is True`, then this superclass will flatten out all
            the parameters in a model and append them as child items.
        """
        super().__init__(data, model, parent)

        if flat:
            for par in flatten(data.parameters):
                pn = ParNode(par, model, parent=self)
                self.appendChild(pn)
        else:
            for p in data.parameters:
                if isinstance(p, Parameters):
                    n = ParametersNode(p, model, self)
                if isinstance(p, Parameter):
                    n = ParNode(p, model, self)
                self.appendChild(n)

    @property
    def component(self):
        return self._data

    def flags(self, column):
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        if column == 0:
            flags |= QtCore.Qt.ItemIsEditable

        # say that you want the Components to be draggable
        flags |= QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled

        return flags

    def data(self, column, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.CheckStateRole:
            return None

        if column > 0:
            return None
        if role == QtCore.Qt.DisplayRole:
            return self._data.name

    def setData(self, column, value, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.CheckStateRole:
            return False

        if column:
            return False

        self.component.name = value
        self._model.dataChanged.emit(self.index, self.index)
        return True


class StructureNode(Node):
    def __init__(self, data, model, parent=QtCore.QModelIndex()):
        super().__init__(data, model, parent)
        for component in data:
            self.appendChild(
                component_class(component)(component, model, self)
            )

    @property
    def structure(self):
        return self._data

    def data(self, column, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.CheckStateRole:
            return None

        if column > 0:
            return None
        # structures name
        return "structure"

    def remove_component(self, row):
        self._model.beginRemoveRows(self.index, row, row)

        # remove all dependent parameters (either in this dataset or elsewhere)
        # this has to be done before it's popped from the Structure.
        self._model.unlink_dependent_parameters(self.child(row))

        self.structure.pop(row)
        self.popChild(row)

        self._model.endRemoveRows()

    def move_component(self, src, dst):
        if src == dst or dst == src + 1:
            return

        self._model.beginMoveRows(self.index, src, src, self.index, dst)

        # swap in the underlying data
        strc = self._data
        children = self._children

        c = strc[src]
        cn = children[src]

        strc.insert(dst, c)
        children.insert(dst, cn)

        if src < dst:
            strc.pop(src)
            children.pop(src)
        else:
            strc.pop(src + 1)
            children.pop(src + 1)

        self._model.endMoveRows()

    def insert_component(self, row, component):
        n = component_class(component)(component, self._model, self)

        self._model.beginInsertRows(self.index, row, row)
        self.insertChild(row, n)
        self.structure.insert(row, component)
        self._model.endInsertRows()


# the child note where structures start appearing in a ReflectModelNode
STRUCT_OFFSET = 4


class ReflectModelNode(Node):
    def __init__(self, data, model, parent=QtCore.QModelIndex()):
        super().__init__(data, model, parent=parent)

        # pointwise/constant dq/q choice is kept in the ReflectModel
        if data.dq_type == "constant":
            self.constantdq_q = True
        elif data.dq_type == "pointwise":
            self.constantdq_q = False

        # deal with scale factors
        if isinstance(data, ReflectModel):
            n = ParNode(data.scale, model, self)
        elif isinstance(data, MixedReflectModel):
            n = ParametersNode(data.scales, model, self)
        self.appendChild(n)

        n = ParNode(data.bkg, model, self)
        self.appendChild(n)
        n = ParNode(data.dq, model, self)
        self.appendChild(n)
        n = ParNode(data.q_offset, model, self)
        self.appendChild(n)

        if isinstance(data, ReflectModel):
            n = StructureNode(data.structure, model, self)
            self.appendChild(n)
        elif isinstance(data, MixedReflectModel):
            for structure in data.structures:
                n = StructureNode(structure, model, self)
                self.appendChild(n)

    @property
    def structures(self):
        return [
            self.child(i)._data
            for i in range(STRUCT_OFFSET, self.childCount())
        ]

    def data(self, column, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.CheckStateRole and column == 1:
            if self.constantdq_q:
                return QtCore.Qt.Checked
            else:
                return QtCore.Qt.Unchecked

        if role == QtCore.Qt.DisplayRole:
            if column == 0:
                # structures name
                return "model"
            elif column == 1:
                return "dq/q const. smearing?"

        return None

    def remove_structure(self, row):
        # don't remove the last structure
        if row < STRUCT_OFFSET or len(self.structures) < 2:
            return

        # you have to be a mixedreflectmodel because there's more than one
        # structure
        data_object_node = find_data_object(self.index)
        data_object = data_object_node.data_object
        orig_model = data_object.model
        structures = orig_model.structures
        scales = orig_model.scales

        # going down to a single reflectmodel
        if len(structures) == 2:
            self._model.beginRemoveRows(self.index, row, row)
            # remove all dependent parameters (either in this dataset or
            # elsewhere)
            # do this before the Structure is popped.
            self._model.unlink_dependent_parameters(self.child(row))

            structures.pop(row - STRUCT_OFFSET)
            scales.pop(row - STRUCT_OFFSET)
            sf = possibly_create_parameter(scales[0], name="scale")
            new_model = ReflectModel(
                structures[0], scale=sf, bkg=orig_model.bkg, dq=orig_model.dq
            )
            data_object.model = new_model
            self.popChild(row)
            self._model.endRemoveRows()
            n = ParNode(sf, self._model, self)
            self._model.beginInsertRows(self.index, 0, 0)
            self.insertChild(0, n)
            self._model.endInsertRows()
            self._model.beginRemoveRows(self.index, 1, 1)
            self.popChild(1)
            self._model.endRemoveRows()
            # data_object_node.set_reflect_model(new_model)
            return

        # you're not down to a single structure, so there must have been more
        # than 2 structures
        self._model.beginRemoveRows(self.index, row, row)

        # remove all dependent parameters (either in this dataset or elsewhere)
        # do this before the Structure is popped.
        self._model.unlink_dependent_parameters(self.child(row))

        # pop the structure and scale factor nodes
        self.popChild(row)
        structures.pop(row - STRUCT_OFFSET)
        scales.pop(row - STRUCT_OFFSET)
        self._model.endRemoveRows()

        self._model.beginRemoveRows(
            self.child(0).index, row - STRUCT_OFFSET, row - STRUCT_OFFSET
        )
        self.child(0).popChild(row - STRUCT_OFFSET)
        self._model.endRemoveRows()

    def insert_structure(self, row, structure):
        n = StructureNode(structure, self._model, self)

        data_object_node = find_data_object(self.index)
        data_object = data_object_node.data_object
        orig_model = data_object.model

        if len(self.structures) == 1:
            self._model.beginInsertRows(self.index, row, row)
            new_structures = [self.structures[0], structure]
            new_model = MixedReflectModel(
                new_structures, bkg=orig_model.bkg, dq=orig_model.dq
            )
            data_object.model = new_model
            data_object_node.set_reflect_model(
                new_model, constdq_q=self.constantdq_q
            )
            return

        # already a mixed model
        # we can't insert at a lower place than the 4rd row
        row = max(row, 4)
        self._model.beginInsertRows(self.index, row, row)

        # insert the structure
        orig_model.structures.insert(row - STRUCT_OFFSET, structure)
        v = 1 / len(orig_model.structures)
        sf = possibly_create_parameter(v, name="scale")
        orig_model.scales.insert(row - STRUCT_OFFSET, sf)

        self.insertChild(row, n)
        self._model.endInsertRows()

        # insert a scale factor
        self._model.beginInsertRows(
            self.child(0).index, row - STRUCT_OFFSET, row - STRUCT_OFFSET
        )
        # add a scale factor
        n = ParNode(sf, self._model, self.child(0))
        self.child(0).insertChild(row - STRUCT_OFFSET, n)
        self._model.endInsertRows()

    def setData(self, column, value, role=QtCore.Qt.EditRole):
        # currently this only deals with constant dq/q
        if role == QtCore.Qt.CheckStateRole and column == 1:
            self.constantdq_q = value == QtCore.Qt.Checked
            data_object_node = find_data_object(self.index)
            data_object = data_object_node.data_object
            data_object.constantdq_q = self.constantdq_q

            # need to not let the dq parameter vary if there's resolution
            # information in the dataset
            if not self.constantdq_q:
                data_object_node.data_object.model.dq_type = "pointwise"
                data_object_node.data_object.model.dq.vary = False
            else:
                data_object_node.data_object.model.dq_type = "constant"

            lindex = self._model.index(2, 0, self.index)
            rindex = self._model.index(2, 4, self.index)
            self._model.dataChanged.emit(lindex, rindex)

            return True

        return True

    def flags(self, column):
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        if column == 1:
            flags |= QtCore.Qt.ItemIsUserCheckable

        return flags


class DatasetNode(Node):
    def __init__(self, data, model, parent=QtCore.QModelIndex()):
        super().__init__(data, model, parent)

    @property
    def dataset(self):
        return self._data

    def refresh(self):
        self.dataset.refresh()

    def flags(self, column):
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        return flags

    def data(self, column, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.CheckStateRole:
            return None

        if role == QtCore.Qt.DisplayRole:
            if column == 0:
                return self._data.name
            return None


class DataObjectNode(Node):
    def __init__(self, data_object, model, parent=QtCore.QModelIndex()):
        super().__init__(data_object, model, parent=parent)

        self.chi2 = float(np.nan)
        self.visible = data_object.graph_properties.visible

        n = DatasetNode(data_object.dataset, model, self)
        self.appendChild(n)

        # append the ReflectModel.
        if data_object.model is not None:
            # already existing
            n = ReflectModelNode(data_object.model, model, self)
            self.appendChild(n)

    @property
    def data_object(self):
        return self._data

    def refresh(self):
        self.child(0).refresh()
        # recalculate number of points
        index = self.parent().index
        idx = self._model.index(self.row(), 1, index)
        idx1 = self._model.index(self.row(), 1, index)

        self._model.dataChanged.emit(idx, idx1)

    def set_reflect_model(self, model, constdq_q=True):
        if model is not None:
            n = ReflectModelNode(model, self._model, self)
            n.constantdq_q = constdq_q

            # no ReflectModel, append the ReflectModel as a child
            if len(self._children) == 1:
                self._model.beginInsertRows(self.index, 1, 1)
                self.appendChild(n)
                self.data_object.model = model
                self.data_object.constantdq_q = constdq_q
                self._model.endInsertRows()
            # there is a pre-existing ReflectModel, overwrite.
            elif len(self._children) == 2:
                # remove and add
                self._model.beginRemoveRows(self.index, 1, 1)
                self.popChild(1)
                self._model.endRemoveRows()
                self._model.beginInsertRows(self.index, 1, 1)
                self.appendChild(n)
                self.data_object.model = model
                self.data_object.constantdq_q = constdq_q
                self._model.endInsertRows()

    def set_dataset(self, dataset):
        if dataset is not None:
            n = DatasetNode(dataset, self._model, parent=self)
            self.data_object.dataset = dataset
            self._children[0] = n
            index = self._children[0].index
            self._model.dataChanged.emit(index, index)

    def columnCount(self):
        return 4

    def setData(self, column, value, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.CheckStateRole and column == 1:
            self.visible = value == QtCore.Qt.Checked
            return True

        return True

    def data(self, column, role=QtCore.Qt.DisplayRole):
        if role == QtCore.Qt.CheckStateRole and column == 1:
            if self.visible:
                return QtCore.Qt.Checked
            else:
                return QtCore.Qt.Unchecked

        if role == QtCore.Qt.DisplayRole:
            if column == 0:
                return self._data.name
            elif column == 1:
                return "display"
            elif column == 2:
                return "points: %d" % len(self._data.dataset)
            elif column == 3:
                return "chi2: %g" % self.chi2
        elif role == QtCore.Qt.ToolTipRole:
            if column == 1:
                return "Show or hide the dataset from the graphs"
            elif column == 3:
                return (
                    "((y<sub>i,data</sub>-y<sub>i,model</sub>)"
                    "/y<sub>i,err</sub>)<sup>2</sup> summed over all "
                    "datapoints"
                )

    def flags(self, column):
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        if column == 1:
            flags |= QtCore.Qt.ItemIsUserCheckable

        # want to drop dataobject onto currently_fitting
        flags |= QtCore.Qt.ItemIsDragEnabled
        return flags


class ContainerNode(Node):
    # set up with a datastore
    def __init__(self, data, model, parent=None):
        super().__init__(data, model, parent)
        for data_object in data:
            n = DataObjectNode(data_object, model, parent=self)
            self.appendChild(n)

    @property
    def datastore(self):
        return self._data

    def load_data(self, filename):
        n_objects = len(self.datastore)
        data_object = self.datastore.load(filename)

        if data_object is not None:
            # if you load a new dataset then add it to the model
            if n_objects - len(self.datastore):
                position = self.childCount()
                self._model.beginInsertRows(self.index, position, position)
                n = DataObjectNode(data_object, self._model, parent=self)
                self.appendChild(n)
                self._model.endInsertRows()
            else:
                # if the number of datasets didn't increase, then it was
                # refreshed.
                position = self.datastore.names.index(data_object.name)
                index = self.index
                idx = self._model.index(position, 1, index)
                idx1 = self._model.index(position, 3, index)

                self._model.dataChanged.emit(idx, idx1)

        return data_object

    def set_data_object(self, data_object):
        if data_object is None:
            return

        # see if it's already present
        if data_object.name in self.datastore.names:
            row = self.datastore.names.index(data_object.name)
            self.child(row).set_dataset(data_object.dataset)
            self.child(row).set_reflect_model(data_object.model)
        else:
            position = self.childCount()
            self._model.beginInsertRows(self.index, position, position)
            self.datastore.data_objects[data_object.name] = data_object
            n = DataObjectNode(data_object, self._model, parent=self)
            self.appendChild(n)
            self._model.endInsertRows()

    def remove_data_object(self, name):
        try:
            row = self.datastore.names.index(name)
        except ValueError:
            return

        self._model.beginRemoveRows(self.index, row, row)

        # remove all dependent parameters (either in this dataset or elsewhere)
        # do this before the data_object is popped.
        self._model.unlink_dependent_parameters(
            self._model.data_object_node(name)
        )

        # remove from the backing datastore and pop from the nodelist
        self.datastore.remove_dataset(name)
        self.popChild(row)

        self._model.endRemoveRows()

    def refresh(self):
        for data_object_node in self._children:
            data_object_node.refresh()


class TreeModel(QtCore.QAbstractItemModel):
    """
    Parameters
    ----------
    data: refnx.reflect._app.datastore.DataStore
    """

    def __init__(self, data, parent=None):
        super().__init__(parent)
        self._rootnode = ContainerNode(data, self)
        self._data = data

    @property
    def datastore(self):
        return self._data

    def columnCount(self, parent):
        if parent.isValid():
            return parent.internalPointer().columnCount()
        else:
            return self._rootnode.columnCount()

    def data(self, index, role):
        if not index.isValid():
            return None

        item = index.internalPointer()
        return item.data(index.column(), role=role)

    def setData(self, index, value, role=QtCore.Qt.EditRole):
        if not index.isValid():
            return None

        item = index.internalPointer()
        col = index.column()

        # devolve setting of data to Node
        ok = item.setData(col, value, role=role)
        if ok:
            self.dataChanged.emit(index, index, [role])

        return ok

    def flags(self, index):
        if not index.isValid():
            return QtCore.Qt.NoItemFlags

        # we want the parameter value, lb, ub values to be editable
        item = index.internalPointer()

        return item.flags(index.column())

    def headerData(self, section, orientation, role):
        headers = ["Name", "value", "sigma", "lb", "ub", "constraint"]
        if (
            orientation == QtCore.Qt.Horizontal
            and role == QtCore.Qt.DisplayRole
        ):
            return headers[section]

        return None

    def index(self, row, column, parent=QtCore.QModelIndex()):
        if parent is None:
            parent = QtCore.QModelIndex()

        if not parent.isValid():
            parentItem = self._rootnode
        else:
            parentItem = parent.internalPointer()

        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QtCore.QModelIndex()

    def parent(self, index):
        if not index.isValid():
            return QtCore.QModelIndex()

        childItem = index.internalPointer()
        parentItem = childItem.parent()

        if parentItem == self._rootnode:
            return QtCore.QModelIndex()

        return self.createIndex(parentItem.row(), 0, parentItem)

    def supportedDropActions(self):
        return QtCore.Qt.MoveAction

    def mimeTypes(self):
        return ["application/vnd.treeviewdragdrop.list"]

    def mimeData(self, indexes):
        index_info = []
        mimedata = QtCore.QMimeData()
        for index in indexes:
            if index.isValid():
                node = index.internalPointer()
                info = {"name": None, "indices": None}
                if isinstance(node, Node):
                    info["indices"] = node.row_indices()
                if isinstance(node, DataObjectNode):
                    info["name"] = node.data_object.name
                index_info.append(info)

        s = pickle.dumps(index_info)
        mimedata.setData("application/vnd.treeviewdragdrop.list", s)
        return mimedata

    def dropMimeData(self, data, action, row, column, parent):
        if action == QtCore.Qt.IgnoreAction:
            return True

        if not data.hasFormat("application/vnd.treeviewdragdrop.list"):
            return False

        # what the destination was.
        host_node = parent.internalPointer()

        if not (
            isinstance(host_node, ComponentNode)
            or isinstance(host_node, StackNode)
        ):
            return False

        # host_structure could be a Structure OR a Stack
        host_structure_node = host_node.parent()
        host_structure = host_structure_node._data
        ba = data.data("application/vnd.treeviewdragdrop.list")
        index_info = pickle.loads(ba)
        dragged_nodes = list(
            unique(
                self.node_from_row_indices(i["indices"]) for i in index_info
            )
        )

        # order the dragged nodes
        src_rows = [dn.row() for dn in dragged_nodes]
        d = sorted(zip(src_rows, dragged_nodes), key=itemgetter(0))
        dragged_nodes = [i[1] for i in d]

        # add to the destination in reverse
        dragged_nodes.reverse()

        for dragged_node in dragged_nodes:
            # can't drag a fronting/backing medium in a Structure
            src_structure_node = dragged_node.parent()
            src_structure = src_structure_node._data

            if isinstance(
                src_structure_node, StructureNode
            ) and dragged_node.row() in [
                0,
                len(src_structure) - 1,
            ]:
                continue

            # if (isinstance(dragged_node, SplineNode) and
            #         isinstance(host_structure_node, StackNode)):
            #     continue

            # figure out what the destination is.
            dst_row = host_node.row() + 1
            if dst_row == len(host_structure) and isinstance(
                host_structure_node, StructureNode
            ):
                return False

            if src_structure_node is host_structure_node:
                # moving within the same structure
                src_row = dragged_node.row()
                host_structure_node.move_component(src_row, dst_row)
            else:
                # move the Component between Structures.
                c = deepcopy(dragged_node._data)
                src_row = dragged_node.row()
                src_structure_node.remove_component(src_row)
                host_structure_node.insert_component(dst_row, c)

        return True

    def rowCount(self, parent=QtCore.QModelIndex()):
        if not parent.isValid():
            parentItem = self._rootnode
        else:
            parentItem = parent.internalPointer()

        return parentItem.childCount()

    def rebuild(self):
        self._rootnode = ContainerNode(self._data, self)
        self.modelReset.emit()

    def load_data(self, filename):
        data_object = self._rootnode.load_data(filename)
        return data_object

    def remove_data_object(self, name):
        self._rootnode.remove_data_object(name)

    def data_object_row(self, name):
        # returns a row in the rootnode given a name
        try:
            return self.datastore.names.index(name)
        except ValueError:
            return None

    def data_object_node(self, name):
        # returns a DataObjectNode given a name
        row = self.data_object_row(name)
        if row is not None:
            return self._rootnode.child(row)
        else:
            return None

    def node_from_row_indices(self, row_indices):
        # retrieve a descendant node given a series of row indices
        # the first entry in row_indices should always be 0/None.
        node = self._rootnode
        for idx in row_indices[1:]:
            node = node.child(idx)

        return node

    def snapshot(self, snapshot_name):
        original = self.datastore["theoretical"]
        dataset = ReflectDataset()
        dataset.data = (
            original.dataset.x,
            original.model.model(original.dataset.x, x_err=dataset.x_err),
        )
        dataset.name = snapshot_name

        new_model = deepcopy(original.model)
        new_model.name = snapshot_name

        # if the snapshot already exists then overwrite it.
        if snapshot_name in self.datastore.names:
            row = self.data_object_row(snapshot_name)
            self._rootnode.child(row).set_dataset(dataset)
            self._rootnode.child(row).set_reflect_model(new_model)
            data_object = self.data_object_node(snapshot_name).data_object
        else:
            # otherwise you have to add it.
            data_object = DataObject(dataset)
            data_object.model = new_model
            self._rootnode.set_data_object(data_object)

        return data_object

    def refresh(self):
        self._rootnode.refresh()

    def unlink_dependent_parameters(self, node):
        # if a dataset, or component, is removed which contains a 'master'
        # parameter, this method iterates through all dataobjects and removes
        # all constraints for parameters (on all datasets) that were linked to
        # those master parameters.

        # all the data_object nodes
        dons = self._rootnode._children

        if isinstance(node, DataObjectNode):
            # removing a dataset
            o_params = node.data_object.model.parameters.flattened()
        elif isinstance(node, ComponentNode):
            o_params = node.component.parameters.flattened()
        elif isinstance(node, StructureNode):
            o_params = node.structure.parameters.flattened()
        elif isinstance(node, StackNode):
            o_params = node.stack.parameters.flattened()

        # iterate through dataobjects and see if any parameters depend on
        # any of these model parameters
        for don in dons:
            do = don.data_object
            model_pars = do.model.parameters.flattened()
            for mp in model_pars:
                if set(mp.dependencies()).intersection(o_params):
                    mp.constraint = None
            # TODO: investigate whether a dataChanged emit is necessary.


class TreeFilter(QtCore.QSortFilterProxyModel):
    def __init__(self, tree_model, parent=None, only_fitted=False):
        super(TreeFilter, self).__init__(parent)
        self.tree_model = tree_model
        # only_fitted means that only the entries that are varying will be
        # displayed.
        self.only_fitted = only_fitted
        self._fitted_datasets = []

    def filterAcceptsRow(self, row, index):
        idx = self.tree_model.index(row, 0, index)

        if not idx.isValid():
            return False

        item = idx.internalPointer()
        if isinstance(item, DatasetNode):
            return False

        # filter out resolution parameter if the dataset has x_err
        # and constant dq/q wasn't requested.
        if isinstance(item.parent(), ReflectModelNode) and isinstance(
            item, ParNode
        ):
            data_object_node = find_data_object(item.index)
            dataset = data_object_node.data_object.dataset
            constantdq_q = data_object_node.data_object.constantdq_q
            # hard-coded the row for dq/q
            if (
                item.row() == 2
                and not constantdq_q
                and dataset.x_err is not None
            ):
                return False

        # filter out parameters for the fronting/backing media
        if isinstance(item.parent(), SlabNode) and isinstance(item, ParNode):

            # component
            parent = item.parent()
            struc = parent.parent()
            if not isinstance(struc, StructureNode):
                return True

            component_loc = parent.row()
            if component_loc == 0 and row in [0, 2, 3, 4]:
                return False
            if component_loc == len(struc.structure) - 1 and row in [0, 4]:
                return False

        # only_fitted means that only varying parameters should be displayed
        if self.only_fitted:
            # you have to be in the currently fitting list to be shown
            dataset = find_data_object(idx).data_object
            if dataset.name not in self._fitted_datasets:
                return False
            if isinstance(item, ParNode) and item._data.vary is False:
                return False

        return True


def find_data_object(index):
    if not index.isValid():
        return

    item = index.internalPointer()
    hierarchy = item.hierarchy()
    data_object_node = [i for i in hierarchy if isinstance(i, DataObjectNode)]
    return data_object_node[0]


###############################################################################
# Classes for the Node structure of different Components

###############################################################################
class SlabNode(ComponentNode):
    def __init__(self, data, model, parent=QtCore.QModelIndex()):
        super().__init__(data, model, parent)


###############################################################################
class LipidLeafletNode(ComponentNode):
    def __init__(self, data, model, parent=QtCore.QModelIndex()):
        super().__init__(data, model, parent)

        prop_node = PropertyNode("reverse_monolayer", model, parent=self)
        self.appendChild(prop_node)


###############################################################################
class SplineNode(ComponentNode):
    def __init__(self, data, model, parent=QtCore.QModelIndex()):
        super().__init__(data, model, parent, flat=False)
        prop_node = PropertyNode("zgrad", model, parent=self)
        self.appendChild(prop_node)

        validator = QtGui.QDoubleValidator()
        validator.setBottom(0)
        prop_node = PropertyNode(
            "microslab_max_thickness",
            model,
            parent=self,
            validators=(validator,),
        )
        prop_node.attribute_type = float
        self.appendChild(prop_node)


###############################################################################
class StackNode(Node):
    def __init__(self, data, model, parent=QtCore.QModelIndex()):
        super().__init__(data, model, parent)

        # append the number of repeats
        pn = ParNode(data.repeats, model, parent=self)
        self.appendChild(pn)

        for component in data:
            self.appendChild(
                component_class(component)(component, model, self)
            )

    @property
    def stack(self):
        return self._data

    def flags(self, column):
        flags = QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled
        if column == 0:
            flags |= QtCore.Qt.ItemIsEditable

        # say that you want the Components to be draggable
        flags |= QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled

        return flags

    def data(self, column, role=QtCore.Qt.EditRole):
        if role == QtCore.Qt.CheckStateRole:
            return None

        if column > 0:
            return None

        if role == QtCore.Qt.DisplayRole:
            return self._data.name

    def remove_component(self, row):
        strc = self._data
        if len(strc) == 1:
            # can't remove the last component in a stack.
            return

        self._model.beginRemoveRows(self.index, row, row)
        # the first row in the GUI model is the repeats parnode, the second is
        # the first component
        self.stack.pop(row - 1)
        self.popChild(row)
        self._model.endRemoveRows()

    def move_component(self, src, dst):
        if src == dst or dst == src + 1:
            return

        self._model.beginMoveRows(self.index, src, src, self.index, dst)

        # swap in the underlying data
        strc = self._data
        children = self._children

        c = strc[src - 1]
        cn = children[src]

        strc.insert(dst - 1, c)
        children.insert(dst, cn)

        if src < dst:
            strc.pop(src - 1)
            children.pop(src)
        else:
            strc.pop(src)
            children.pop(src + 1)

        self._model.endMoveRows()

    def insert_component(self, row, component):
        n = component_class(component)(component, self._model, self)

        self._model.beginInsertRows(self.index, row, row)
        self.insertChild(row, n)
        self.stack.insert(row - 1, component)
        self._model.endInsertRows()



import os.path

import pytest

from refnx.dataset import ReflectDataset, Data1D
import numpy as np
from numpy.testing import assert_equal, assert_


class TestReflectDataset:
    @pytest.fixture(autouse=True)
    def setup_method(self, tmpdir):
        self.pth = os.path.dirname(os.path.abspath(__file__))

        data = ReflectDataset()

        x1 = np.linspace(0, 10, 115)
        y1 = 2 * x1
        e1 = np.ones_like(x1)
        dx1 = np.ones_like(x1)
        data.add_data((x1, y1, e1, dx1))
        self.data = data

        self.cwd = os.getcwd()
        self.tmpdir = tmpdir.strpath
        os.chdir(self.tmpdir)

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_load(self):
        # load dataset from XML, via file handle
        dataset = ReflectDataset()
        with open(os.path.join(self.pth, "c_PLP0000708.xml")) as f:
            dataset.load(f)

        assert_equal(len(dataset), 90)
        assert_equal(90, np.size(dataset.x))

        # load dataset from XML, via string
        dataset = ReflectDataset()
        dataset.load(os.path.join(self.pth, "c_PLP0000708.xml"))

        assert_equal(len(dataset), 90)
        assert_equal(90, np.size(dataset.x))

        # load dataset from XML, via string
        dataset = ReflectDataset(os.path.join(self.pth, "c_PLP0000708.xml"))

        assert_equal(len(dataset), 90)
        assert_equal(90, np.size(dataset.x))

        # load dataset from .dat, via file handle
        dataset1 = ReflectDataset()
        with open(os.path.join(self.pth, "c_PLP0000708.dat")) as f:
            dataset1.load(f)

        assert_equal(len(dataset1), 90)
        assert_equal(90, np.size(dataset1.x))

        # load dataset from .dat, via string
        dataset2 = ReflectDataset()
        dataset2.load(os.path.join(self.pth, "c_PLP0000708.dat"))

        assert_equal(len(dataset2), 90)
        assert_equal(90, np.size(dataset2.x))

    def test_dot(self):
        # test with file formats from http://www.reflectometry.net/refdata.htm
        dataset1 = ReflectDataset()
        with open(os.path.join(self.pth, "dot.dat")) as f:
            dataset1.load(f)
        assert_equal(len(dataset1), 31)
        assert dataset1.y_err is not None
        assert dataset1.x_err is None

        with open(os.path.join(self.pth, "dot.aft")) as f:
            dataset1.load(f)
        assert_equal(len(dataset1), 36)
        assert dataset1.y_err is not None

        with open(os.path.join(self.pth, "dot.refl")) as f:
            dataset1.load(f)
        assert_equal(len(dataset1), 59)
        assert dataset1.y_err is not None

        with open(os.path.join(self.pth, "dot.out")) as f:
            dataset1.load(f)
        assert_equal(len(dataset1), 1122)
        assert dataset1.y_err is not None
        assert dataset1.x_err is not None

        with open(os.path.join(self.pth, "dot.mft")) as f:
            dataset1.load(f)
        assert_equal(len(dataset1), 213)
        assert dataset1.y_err is not None
        assert dataset1.x_err is not None

    def test_load_dat_with_header(self):
        # check that the file load works with a header
        a = ReflectDataset(os.path.join(self.pth, "c_PLP0000708.dat"))
        b = ReflectDataset(os.path.join(self.pth, "c_PLP0000708_header.dat"))
        c = ReflectDataset(os.path.join(self.pth, "c_PLP0000708_header2.dat"))
        assert_equal(len(a), len(b))
        assert_equal(len(a), len(c))

    def test_GH236(self):
        a = ReflectDataset(os.path.join(self.pth, "c_PLP0033831.txt"))
        assert_equal(len(a), 166)

    def test_loading_junk(self):
        # if you can't load anything from a datafile then you should get a
        # RuntimeError raised.
        from pytest import raises

        with raises(RuntimeError):
            ReflectDataset(os.path.join(self.pth, "../__init__.py"))

    def test_construction(self):
        # test we can construct a dataset directly from a file.
        pth = os.path.join(self.pth, "c_PLP0000708.xml")

        ReflectDataset(pth)

        with open(os.path.join(self.pth, "c_PLP0000708.xml")) as f:
            ReflectDataset(f)

        ReflectDataset(os.path.join(self.pth, "d_a.txt"))

    def test_init_with_data1d(self):
        # test we can construct a dataset from a dataset (i.e. a copy)
        dataset = Data1D(self.data)
        assert_equal(dataset.y, self.data.y)

    def test_add_data(self):
        # test we can add data to the dataset

        # 2 columns
        data = Data1D()

        x1 = np.linspace(0, 10, 5)
        y1 = 2 * x1
        data.add_data((x1, y1))

        x2 = np.linspace(7, 20, 10)
        y2 = 2 * x2 + 2
        data.add_data((x2, y2), requires_splice=True)

        assert len(data) == 13

        # 3 columns
        data = Data1D()

        x1 = np.linspace(0, 10, 5)
        y1 = 2 * x1
        e1 = np.ones_like(x1)
        data.add_data((x1, y1, e1))

        x2 = np.linspace(7, 20, 10)
        y2 = 2 * x2 + 2
        e2 = np.ones_like(x2)
        data.add_data((x2, y2, e2), requires_splice=True)

        assert len(data) == 13

        # 4 columns
        data = Data1D()

        x1 = np.linspace(0, 10, 5)
        y1 = 2 * x1
        e1 = np.ones_like(x1)
        dx1 = np.ones_like(x1)
        data.add_data((x1, y1, e1, dx1))

        x2 = np.linspace(7, 20, 10)
        y2 = 2 * x2 + 2
        e2 = np.ones_like(x2)
        dx2 = np.ones_like(x2)

        data.add_data((x2, y2, e2, dx2), requires_splice=True)

        assert len(data) == 13

        # test addition of datasets.
        x1 = np.linspace(0, 10, 5)
        y1 = 2 * x1
        e1 = np.ones_like(x1)
        dx1 = np.ones_like(x1)
        data = Data1D((x1, y1, e1, dx1))

        x2 = np.linspace(7, 20, 10)
        y2 = 2 * x2 + 2
        e2 = np.ones_like(x2)
        dx2 = np.ones_like(x2)
        data2 = Data1D((x2, y2, e2, dx2))

        data3 = data + data2
        assert len(data3) == 13

        # test iadd of datasets
        data += data2
        assert len(data) == 13

    def test_add_non_overlapping(self):
        # check that splicing of non-overlapping datasets raises ValueError
        x1 = np.linspace(0, 10, 11)
        y1 = 2 * x1
        data = Data1D((x1, y1))

        x2 = np.linspace(11, 20, 10)
        y2 = 2 * x2 + 2

        from pytest import raises

        with raises(ValueError):
            data.add_data((x2, y2), requires_splice=True)

    def test_save_xml(self):
        self.data.save_xml("test.xml")
        with open("test.xml", "wb") as f:
            self.data.save_xml(f)

    def test_data_setter(self):
        # check that setting a Data1D's data from a tuple works correctly
        # this is the approach used in Data1D.synthesise.
        new_dataset = Data1D()
        new_dataset.data = self.data.data
        assert_equal(new_dataset.y_err, self.data.y_err)
        assert_equal(new_dataset.x_err, self.data.x_err)
        assert_equal(new_dataset.y, self.data.y)
        assert_equal(new_dataset.x, self.data.x)
        assert_equal(new_dataset.weighted, self.data.weighted)

    def test_synthesise(self):
        # add gaussian noise to a Data1D.y
        new_dataset = self.data.synthesise(random_state=1)

        # y-array should not be the same, so don't test that
        assert_equal(new_dataset.y_err, self.data.y_err)
        assert_equal(new_dataset.x_err, self.data.x_err)
        assert_equal(new_dataset.x, self.data.x)
        assert_equal(new_dataset.weighted, self.data.weighted)

        # synthesis should be repeatable with provision of a seed
        new_dataset2 = self.data.synthesise(random_state=1)
        assert_equal(new_dataset.y, new_dataset2.y)

    def test_mask(self):
        # if you mask all points there should be none left
        self.data.mask = np.full_like(self.data.y, False, bool)
        assert_equal(len(self.data), 0)

        # try masking a random selection
        rando = np.random.randint(0, 2, self.data._y.size)
        self.data.mask = rando
        assert_equal(len(self.data), np.count_nonzero(rando))

        # now clear
        self.data.mask = None
        assert_equal(len(self.data), self.data._y.size)

    def test_repr(self):
        a = Data1D(os.path.join(self.pth, "c_PLP0033831.txt"))
        b = eval(repr(a))
        assert_equal(len(a), 166)
        assert_equal(len(b), 166)

        # load dataset from XML, via string
        a = ReflectDataset(os.path.join(self.pth, "c_PLP0000708.xml"))
        b = eval(repr(a))
        assert_equal(len(b), len(a))
        assert_equal(len(b), len(a))


class TestData1D:
    @pytest.fixture(autouse=True)
    def setup_method(self, tmpdir):
        self.pth = os.path.dirname(os.path.abspath(__file__))
        self.cwd = os.getcwd()
        self.tmpdir = tmpdir.strpath
        os.chdir(self.tmpdir)

    def teardown_method(self):
        os.chdir(self.cwd)

    def test_2D_data(self):
        # despite the class being called Data1D it should be possible to
        # store ND data in it

        # try with multidimensional x data
        x = np.linspace(0.01, 0.5, 100).reshape(50, 2)
        y = np.arange(50, dtype=float)

        data = Data1D((x, y))
        assert len(data) == 50

        assert_equal(data.x, x)
        assert_equal(data.y, y)

        mask = np.ones_like(y, bool)
        mask[1] = False
        data.mask = mask
        assert len(data) == 49
        assert_equal(data.y, np.r_[y[0], y[2:]])
        assert_equal(data.x, np.concatenate((x[0][np.newaxis, :], x[2:])))

        # try with storing multiple channels for y as well
        # this might happen with something like ellipsometry data where you'd
        # have (delta, xsi) as the y data, and (lambda, aoi) as the x data
        # NOTE THAT MASKING WON'T WORK, FOR ND YDATA. This is because the class
        # is using fancy indexing with booleans to remove entries that are
        # unwanted. Could mark as NaN instead?
        x = np.linspace(0.01, 0.5, 50).reshape(50)
        y = np.arange(100, dtype=float).reshape(50, 2)

        data = Data1D((x, y))
        assert len(data) == 100

        assert_equal(data.x, x)
        assert_equal(data.y, y)

        x = np.linspace(0.01, 0.5, 100).reshape(50, 2)

        data = Data1D((x, y))
        assert len(data) == 100

        assert_equal(data.x, x)
        assert_equal(data.y, y)



from refnx.util.ErrorProp import (
    EPadd,
    EPsub,
    EPdiv,
    EPmul,
    EPcos,
    EPexp,
    EPlog,
    EPlog10,
    EPmulk,
    EPpow,
    EPpowk,
    EPsin,
    EPtan,
)
from refnx.util.nsplice import get_scaling_in_overlap
from refnx.util.general import (
    div,
    q,
    q2,
    qcrit,
    double_chopper_frequency,
    wavelength,
    angle,
    tauC,
    transmission_double_chopper,
    transmission_single_chopper,
    penetration_depth,
    resolution_double_chopper,
    resolution_single_chopper,
    slit_optimiser,
    beamfrac,
    beamfrackernel,
    height_of_beam_after_dx,
    energy_wavelength,
    xray_energy,
    xray_wavelength,
    velocity_wavelength,
    wavelength_velocity,
    wavelength_energy,
    actual_footprint,
    neutron_transmission,
)
from refnx.util.quickplot import refplot

from refnx._lib._testutils import PytestTester

test = PytestTester(__name__)
del PytestTester


__all__ = [s for s in dir() if not s.startswith("_")]
